# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "11/10/2021"


from typing import Optional
from silx.gui import qt
from tomwer.core.settings import SlurmSettings


class SlurmSettingsDialog(qt.QDialog):
    sigConfigChanged = qt.Signal()
    """Signal emit when the SlurmSetting changed"""

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self.setLayout(qt.QVBoxLayout())
        self._mainWidget = SlurmSettingsWidget(parent=self)
        self.layout().addWidget(self._mainWidget)

        # buttons for validation
        self._buttons = qt.QDialogButtonBox(self)
        types = qt.QDialogButtonBox.Close
        self._buttons.setStandardButtons(types)
        self.layout().addWidget(self._buttons)

        self._buttons.button(qt.QDialogButtonBox.Close).clicked.connect(self.close)

        # connect signal /slot
        self._mainWidget.sigConfigChanged.connect(self._configChanged)

    def _configChanged(self, *args, **kwargs):
        self.sigConfigChanged.emit()

    def isSlurmActive(self):
        return self._mainWidget.isSlurmActive()

    def getConfiguration(self) -> dict:
        return self._mainWidget.getConfiguration()

    def setConfiguration(self, config: dict) -> None:
        self._mainWidget.setConfiguration(config=config)


class SlurmSettingsWidget(qt.QWidget):
    """Widget used to define Slurm configuration to be used"""

    sigConfigChanged = qt.Signal()
    """Signal emit when the SlurmSetting changed"""

    def __init__(self, parent=None) -> None:
        super().__init__(parent=parent)
        self.setLayout(qt.QFormLayout())

        # queues
        self._queue = qt.QLineEdit("", self)
        self.layout().addRow("queue / partition", self._queue)
        # TODO: add the result of a sinfo retrieving available queues

        # n workers
        self._nWorkers = qt.QSpinBox(self)
        self._nWorkers.setRange(1, 100)
        self.layout().addRow("number of workers", self._nWorkers)

        # ncores active
        self._nCores = qt.QSpinBox(self)
        self._nCores.setRange(1, 500)
        self.layout().addRow("number of cores per worker", self._nCores)

        # memory
        self._memory = qt.QSpinBox(self)
        self._memory.setRange(1, 10000)
        self._memory.setSuffix("GB")
        self.layout().addRow("memory per worker", self._memory)

        # gpu
        self._nGpu = qt.QSpinBox(self)
        self._nGpu.setRange(1, 10)
        self.layout().addRow("number of GPUs per worker", self._nGpu)

        # wall time
        self._wallTimeQLE = qt.QLineEdit("", self)
        self._wallTimeLabel = qt.QLabel("wall time", self)
        self.layout().addRow(self._wallTimeLabel, self._wallTimeQLE)

        # python exe
        self._pythonVenv = qt.QLineEdit("", self)
        self.layout().addRow("python venv", self._pythonVenv)
        self._pythonVenv.setToolTip(
            """
            python virtual environment path to be used to call nabu for reconstructing slices.
            """
        )

        # project name
        self._projectName = qt.QLineEdit("", self)
        self._projectName.setToolTip(
            """Name of the project. Can contains several keyword that will be formatted:
            - scan: will be replaced by the scan name
            - process: will be replaced by the process name (nabu slices, nabu volume...)
            - info: some extra information that can be provided by the process to `specify` the processing
            note: those key word have to be surronding by surrounded `{` and `}` in order to be formatted.
            """
        )
        self._project_nameQLabel = qt.QLabel("project name", self)
        self.layout().addRow(self._project_nameQLabel, self._projectName)

        # port
        # TODO: replace by a dedicated widget / validator for TCP adress
        self._port = _PortRangeSelection(self)
        self._port.setRange(0, 99999)
        self._port.setToolTip("TCP Port for the dask-distributed scheduler")
        self._portLabel = qt.QLabel("port", self)
        self.layout().addRow(self._portLabel, self._port)

        # dashboard port
        self._dashboardPort = qt.QSpinBox(self)
        self._dashboardPort.setRange(0, 99999)
        self._dashboardPort.setToolTip("TCP Port for the dashboard")
        self._dashboardPortLabel = qt.QLabel("dashboard port", self)
        self.layout().addRow(self._dashboardPortLabel, self._dashboardPort)

        # simplify gui
        self._wallTimeLabel.hide()
        self._wallTimeQLE.hide()
        self._dashboardPort.hide()
        self._dashboardPortLabel.hide()
        self._projectName.hide()
        self._project_nameQLabel.hide()
        self._portLabel.hide()  # for now we don't use the port. This can be done automatically
        self._port.hide()  # for now we don't use the port. This can be done automatically

        # set up the gui
        self._nCores.setValue(SlurmSettings.N_CORES_PER_WORKER)
        self._nWorkers.setValue(SlurmSettings.N_WORKERS)
        self._memory.setValue(SlurmSettings.MEMORY_PER_WORKER)
        self._queue.setText(SlurmSettings.QUEUE)
        self._nGpu.setValue(SlurmSettings.N_GPUS_PER_WORKER)
        self._projectName.setText(SlurmSettings.PROJECT_NAME)
        self._wallTimeQLE.setText(SlurmSettings.DEFAULT_WALLTIME)
        self._pythonVenv.setText(SlurmSettings.PYTHON_VENV)
        self._port.setRange(*SlurmSettings.PORT_RANGE)
        self._dashboardPort.setValue(SlurmSettings.DASHBOARD_PORT)

        # connect signal / slot
        self._nCores.valueChanged.connect(self._configurationChanged)
        self._nWorkers.valueChanged.connect(self._configurationChanged)
        self._memory.valueChanged.connect(self._configurationChanged)
        self._queue.textEdited.connect(self._configurationChanged)
        self._nGpu.valueChanged.connect(self._configurationChanged)
        self._projectName.editingFinished.connect(self._configurationChanged)
        self._wallTimeQLE.editingFinished.connect(self._configurationChanged)
        self._pythonVenv.editingFinished.connect(self._configurationChanged)
        self._port.sigRangeChanged.connect(self._configurationChanged)
        self._dashboardPort.valueChanged.connect(self._configurationChanged)

    def _configurationChanged(self, *args, **kwargs):
        self.sigConfigChanged.emit()

    def getNCores(self) -> int:
        return self._nCores.value()

    def setNCores(self, n: int) -> None:
        self._nCores.setValue(n)

    def getNWorkers(self) -> int:
        return self._nWorkers.value()

    def setNWorkers(self, n) -> None:
        self._nWorkers.setValue(n)

    def getMemory(self) -> int:
        return self._memory.value()

    def setMemory(self, memory: int) -> None:
        self._memory.setValue(memory)

    def getQueue(self) -> str:
        return self._queue.text()

    def setQueue(self, text: str) -> None:
        self._queue.setText(text)

    def getNGPU(self) -> int:
        return self._nGpu.value()

    def setNGPU(self, n: int) -> None:
        self._nGpu.setValue(n)

    def getProjectName(self):
        return self._projectName.text()

    def setProjectName(self, name):
        self._projectName.setText(name)

    def getWallTime(self):
        return self._wallTimeQLE.text()

    def setWallTime(self, walltime):
        self._wallTimeQLE.setText(walltime)

    def getPythonExe(self):
        return self._pythonVenv.text()

    def setPythonExe(self, python_venv):
        self._pythonVenv.setText(python_venv)

    def getPortRange(self) -> tuple:
        return self._port.getRange()

    def setPortRange(self, min_: int, max_: int, step) -> None:
        self._port.setRange(min_, max_, step)

    def getDashboardPort(self):
        return self._dashboardPort.value()

    def setDashboardPort(self, value):
        self._dashboardPort.setValue(value)

    def setConfiguration(self, config: dict) -> None:
        old = self.blockSignals(True)
        active_slurm = config.get("active_slurm", None)
        if active_slurm is not None:
            self._slurmCB.setChecked(active_slurm)

        n_cores = config.get("cores", None)
        if n_cores is not None:
            self.setNCores(n_cores)

        n_workers = config.get("n_workers", None)
        if n_workers is not None:
            self.setNWorkers(n_workers)

        memory = config.get("memory", None)
        if memory is not None:
            self.setMemory(memory)

        queue_ = config.get("queue", None)
        if queue_ is not None:
            self.setQueue(queue_)

        n_gpu = config.get("n_gpus", None)
        if n_gpu is not None:
            self.setNGPU(n_gpu)

        project_name = config.get("project_name")
        if project_name is not None:
            self.setProjectName(project_name)

        wall_time = config.get("walltime")
        if wall_time is not None:
            self.setWallTime(wall_time)

        python_venv = config.get("python_venv")
        if python_venv is not None:
            self.setPythonExe(python_venv)

        port_range = config.get("port_range")
        if port_range is not None:
            self.setPortRange(*port_range)

        dashboard_port = config.get("dashboard_port")
        if dashboard_port is not None:
            self.setDashboardPort(dashboard_port)

        self.blockSignals(old)
        self.sigConfigChanged.emit()

    def getConfiguration(self) -> dict:
        return {
            "cores": self.getNCores(),
            "n_workers": self.getNWorkers(),
            "memory": self.getMemory(),
            "queue": self.getQueue(),
            "n_gpus": self.getNGPU(),
            "project_name": self.getProjectName(),
            "walltime": self.getWallTime(),
            "python_venv": self.getPythonExe(),
            "port_range": self.getPortRange(),
            "dashboard_port": self.getDashboardPort(),
        }

    def getSlurmClusterConfiguration(self):
        from tomwer.core.cluster import SlurmClusterConfiguration

        return SlurmClusterConfiguration().from_dict(self.getConfiguration())


class _PortRangeSelection(qt.QWidget):
    sigRangeChanged = qt.Signal()
    """Signal emit when the port range change"""

    def __init__(self, parent=None) -> None:
        super().__init__(parent=parent)
        self.setLayout(qt.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        # from label
        self._fromLabel = qt.QLabel("from", self)
        self._fromLabel.setAlignment(qt.Qt.AlignRight | qt.Qt.AlignVCenter)
        self.layout().addWidget(self._fromLabel)
        self._fromQSpinBox = qt.QSpinBox(self)
        self.layout().addWidget(self._fromQSpinBox)
        # to label
        self._toLabel = qt.QLabel("to", self)
        self._toLabel.setAlignment(qt.Qt.AlignRight | qt.Qt.AlignVCenter)
        self.layout().addWidget(self._toLabel)
        self._toQSpinBox = qt.QSpinBox(self)
        self.layout().addWidget(self._toQSpinBox)
        # steps label
        self._stepLabel = qt.QLabel("step", self)
        self._stepLabel.setAlignment(qt.Qt.AlignRight | qt.Qt.AlignVCenter)
        self.layout().addWidget(self._stepLabel)
        self._stepQSpinBox = qt.QSpinBox(self)
        self.layout().addWidget(self._stepQSpinBox)

        # connect signal / slot
        self._fromQSpinBox.valueChanged.connect(self._rangeChanged)
        self._toQSpinBox.valueChanged.connect(self._rangeChanged)
        self._stepQSpinBox.valueChanged.connect(self._rangeChanged)

    def _rangeChanged(self, *args, **kwargs):
        self.sigRangeChanged.emit()

    def getRange(self) -> tuple:
        return (
            self._fromQSpinBox.value(),
            self._toQSpinBox.value(),
            self._stepQSpinBox.value(),
        )

    def setRange(self, min_: int, max_: int, step: Optional[int] = None) -> None:
        self._fromQSpinBox.setValue(min(min_, max_))
        self._toQSpinBox.setValue(max(min_, max_))
        if step is not None:
            self._stepQSpinBox.setValue(step)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "27/03/2020"


import logging
import os
import shutil
import tempfile
from silx.gui import qt
from orangecontrib.tomwer.widgets.reconstruction.FtseriesOW import FtseriesOW
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.synctools.ftseries import QReconsParams
from tomwer.test.utils import UtilsTest
from silx.gui.utils.testutils import TestCaseQt
from tomoscan.scanbase import _FOV
import h5py
import gc

logging.disable(logging.INFO)


class TestPyHstWidget(TestCaseQt):
    """class testing the DarkRefWidget"""

    def setUp(self):
        TestCaseQt.setUp(self)
        self._recons_params = QReconsParams()
        self.widget = FtseriesOW(parent=None, _connect_handler=False)
        self.scan_dir = tempfile.mkdtemp()
        # create dataset
        self.master_file = os.path.join(self.scan_dir, "frm_edftomomill_twoentries.nx")
        shutil.copyfile(
            UtilsTest.getH5Dataset(folderID="frm_edftomomill_twoentries.nx"),
            self.master_file,
        )
        self.scan = HDF5TomoScan(scan=self.master_file, entry="entry0000")

    def tearDown(self):
        self._recons_params = None
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        self.qapp.processEvents()
        gc.collect()
        TestCaseQt.tearDown(self)

    def patch_fov(self, value: str):
        with h5py.File(self.scan.master_file, mode="a") as h5s:
            for entry in ("entry0000", "entry0001"):
                entry_node = h5s[entry]
                if "instrument/detector/field_of_view" in entry_node:
                    del entry_node["instrument/detector/field_of_view"]
                entry_node["instrument/detector/field_of_view"] = value

    def testSetConfiguration(self):
        """Make sure the configuration evolve from scan information"""
        axis_widget = self.widget._ftserie._recParamSetEditor._axisWidget
        self.assertFalse(axis_widget._qcbHalfAcq.isChecked())
        self.patch_fov(value=_FOV.HALF.value)
        self.widget._ftserie.setScan(self.scan)
        self.qapp.processEvents()
        self.assertTrue(axis_widget._qcbHalfAcq.isChecked())

        self.patch_fov(value=_FOV.FULL.value)
        self.scan.clear_caches()
        self.qapp.processEvents()
        self.widget._ftserie.setScan(self.scan)
        self.qapp.processEvents()
        self.assertFalse(axis_widget._qcbHalfAcq.isChecked())



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "09/02/2018"


import logging
import os
import shutil
import tempfile
import unittest
from xml.etree import ElementTree
import fabio
import numpy
import h5py
from silx.gui.utils.testutils import TestCaseQt
from tomwer.core.process.reconstruction.darkref.darkrefs import DarkRefs
from tomwer.core.process.reconstruction.darkref.settings import (
    DARKHST_PREFIX,
    REFHST_PREFIX,
)
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core import utils
from tomwer.core.process.reconstruction.darkref.darkrefscopy import DarkRefsCopy
from tomwer.test.utils import UtilsTest, rebaseAcquisition
from tomwer.core.utils.scanutils import MockHDF5
from silx.io.utils import h5py_read_dataset

logging.disable(logging.INFO)


class TestRefCopyEDF(TestCaseQt):
    """
    Test that RefCopy process is correct for acquisition using EDF
    """

    def setUp(self):
        TestCaseQt.setUp(self)
        self.top_dir = tempfile.mkdtemp()
        dataset_with_ref = "test10"
        data_test_dir = UtilsTest.getEDFDataset(dataset_with_ref)
        self.folder_with_ref = os.path.join(self.top_dir, dataset_with_ref)
        shutil.copytree(data_test_dir, self.folder_with_ref)

        dataset_with_no_ref = "test101"
        self.folder_without_ref = os.path.join(self.top_dir, dataset_with_no_ref)
        rebaseAcquisition(src=self.folder_with_ref, dst=self.folder_without_ref)
        [
            os.remove(f)
            for f in DarkRefs.getRefHSTFiles(
                self.folder_without_ref, prefix=REFHST_PREFIX
            )
        ]
        [
            os.remove(f)
            for f in DarkRefs.getDarkHSTFiles(
                self.folder_without_ref, prefix=DARKHST_PREFIX
            )
        ]

        for acqui in (self.folder_with_ref, self.folder_without_ref):
            for _file in os.listdir(acqui):
                if _file.startswith(("refHST", "dark.edf", "darkHST")):
                    os.remove(os.path.join(acqui, _file))
                if acqui == self.folder_without_ref and _file.startswith("ref"):
                    os.remove(os.path.join(acqui, _file))
        # remove darkend file to avoid creation of dark.edf from it
        os.remove(os.path.join(self.folder_without_ref, "darkend0000.edf"))

        # create back the dark.edf
        darkProcess = DarkRefs(
            inputs={
                "ref_folder": self.folder_with_ref,
                "force_sync": True,
                "data": self.folder_with_ref,
            }
        )
        darkProcess.run()
        assert "dark.edf" in os.listdir(self.folder_with_ref)
        assert "refHST0000.edf" in os.listdir(self.folder_with_ref)
        assert "refHST0020.edf" in os.listdir(self.folder_with_ref)
        self._save_dir = tempfile.mkdtemp()

    def tearDown(self):
        for d in (self.top_dir, self._save_dir):
            shutil.rmtree(d)
        self.ref_copy_process = None
        TestCaseQt.tearDown(self)

    def testAutoMode(self):
        """Check behavior of the auto mode"""
        ref_copy_process = DarkRefsCopy(
            inputs={
                "data": EDFTomoScan(self.folder_with_ref),
                "save_dir": self._save_dir,
                "force_sync": True,
                "mode_auto": True,
            }
        )
        ref_copy_process.run()
        # make sure the refCopy is correctly initialized
        self.assertTrue(ref_copy_process.has_flat_stored() is True)
        self.assertTrue(ref_copy_process.has_dark_stored() is True)
        # check process is doing the job
        ref_copy_process = DarkRefsCopy(
            inputs={
                "data": EDFTomoScan(self.folder_without_ref),
                "save_dir": self._save_dir,
                "force_sync": True,
                "mode_auto": True,
            }
        )
        ref_copy_process.run()
        self.assertTrue(
            len(DarkRefs.getRefHSTFiles(self.folder_without_ref, prefix=REFHST_PREFIX))
            > 0
        )
        self.assertTrue(
            len(
                DarkRefs.getDarkHSTFiles(self.folder_without_ref, prefix=DARKHST_PREFIX)
            )
            > 0
        )

    def testManualMode(self):
        """Check behavior of the manual mode"""
        ref_copy_process = DarkRefsCopy(
            inputs={
                "data": EDFTomoScan(self.folder_without_ref),
                "save_dir": self._save_dir,
                "mode_auto": False,
                "init_dkrf_from": EDFTomoScan(self.folder_with_ref),
            }
        )
        # make sure the refCopy is correctly initialized
        self.assertTrue(ref_copy_process.has_dark_stored() is True)
        with h5py.File(ref_copy_process.get_dark_save_file(), "r") as h5f:
            assert "dark_basename" in h5f
            self.assertEqual("dark.edf", h5py_read_dataset(h5f["dark_basename"]))
        self.assertTrue(ref_copy_process.has_flat_stored())
        # check process is doing the job
        ref_copy_process.run()
        self.assertTrue(
            len(DarkRefs.getRefHSTFiles(self.folder_without_ref, prefix=REFHST_PREFIX))
            > 0
        )
        self.assertTrue(
            len(
                DarkRefs.getDarkHSTFiles(self.folder_without_ref, prefix=DARKHST_PREFIX)
            )
            > 0
        )

    def testNormalizationSRCurrent(self):
        """Make sure the srCurrent is taking into account"""
        ref_copy_process = DarkRefsCopy(
            inputs={
                "mode_auto": False,
                "init_dkrf_from": EDFTomoScan(self.folder_with_ref),
                "data": EDFTomoScan(self.folder_without_ref),
                "save_dir": self._save_dir,
            }
        )

        self.assertTrue(os.path.isfile(os.path.join(self.folder_with_ref, "dark.edf")))
        original_ref_data = fabio.open(
            os.path.join(self.folder_with_ref, "refHST0000.edf")
        ).data
        assert utils.getSRCurrent(scan_dir=self.folder_with_ref, when="start") == 101.3
        assert utils.getSRCurrent(scan_dir=self.folder_with_ref, when="end") == 93.6
        with h5py.File(ref_copy_process.get_ref_save_file(), "r", swmr=True) as h5f:
            os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"
            norm_ref_data = h5py_read_dataset(h5f["data_start"])
        self.assertFalse(numpy.array_equal(original_ref_data, norm_ref_data))

        # change SRCurrent on the .xml
        infoXml = os.path.join(self.folder_without_ref, "test101.xml")
        assert os.path.exists(infoXml)
        xmlTree = ElementTree.parse(infoXml)
        xmlTree.getroot()[0][6].text = "200.36"
        xmlTree.getroot()[0][7].text = "199.63"
        xmlTree.write(infoXml)
        ref_copy_process.run()
        self.assertTrue(
            os.path.isfile(os.path.join(self.folder_without_ref, "dark.edf"))
        )
        self.assertTrue(
            os.path.isfile(os.path.join(self.folder_without_ref, "refHST0000.edf"))
        )
        copyRefData = fabio.open(
            os.path.join(self.folder_without_ref, "refHST0000.edf")
        ).data
        assert (
            utils.getSRCurrent(scan_dir=self.folder_without_ref, when="start") == 200.36
        )
        assert (
            utils.getSRCurrent(scan_dir=self.folder_without_ref, when="end") == 199.63
        )
        self.assertFalse(numpy.array_equal(original_ref_data, copyRefData))
        self.assertFalse(numpy.array_equal(norm_ref_data, copyRefData))

    def testInverseOperation(self):
        """Make sure we will found back the original flat field reference
        if the process is forced to take the default value for SRCurrent
        two time"""
        ref_copy_process = DarkRefsCopy(
            inputs={
                "mode_auto": False,
                "init_dkrf_from": EDFTomoScan(self.folder_with_ref),
                "data": EDFTomoScan(self.folder_without_ref),
                "save_dir": self._save_dir,
            }
        )
        originalRefData = fabio.open(
            os.path.join(self.folder_with_ref, "refHST0000.edf")
        ).data
        ref_copy_process.run()
        copyRefData = fabio.open(
            os.path.join(self.folder_without_ref, "refHST0000.edf")
        ).data
        self.assertTrue(originalRefData.shape == copyRefData.shape)
        self.assertTrue(numpy.allclose(originalRefData, copyRefData))


class TestRefCopyHDF5(unittest.TestCase):
    """
    Test that RefCopy process is correct for acquisition using HDF5
    """

    def setUp(self) -> None:
        self._folder = tempfile.mkdtemp()
        self._save_dir = tempfile.mkdtemp()
        mock_with_refs = MockHDF5(
            scan_path=os.path.join(self._folder, "with_refs"),
            n_proj=10,
            n_ini_proj=10,
            create_ini_ref=True,
            create_ini_dark=True,
        )
        self._acquisition_with_refs = mock_with_refs.scan

        mock_without_refs = MockHDF5(
            scan_path=os.path.join(self._folder, "without_refs"),
            n_proj=10,
            n_ini_proj=10,
            create_ini_ref=False,
            create_ini_dark=False,
        )

        self._acquisition_without_refs = mock_without_refs.scan

        # process dark ref to create dark and ref
        tomwer_processes_file = os.path.join(
            self._acquisition_with_refs.path,
            os.path.basename(self._acquisition_with_refs.path) + "_tomwer_processes.h5",
        )
        assert not os.path.exists(tomwer_processes_file)
        dkref_process = DarkRefs(
            inputs={
                "force_sync": True,
                "data": self._acquisition_with_refs,
            }
        )
        dkref_process.run()
        assert os.path.exists(tomwer_processes_file)
        self.darks = DarkRefs.get_darks_frm_process_file(tomwer_processes_file)
        self.flats = DarkRefs.get_flats_frm_process_file(tomwer_processes_file)
        assert len(self.darks) == 1
        assert 0 in self.darks
        assert len(self.flats) == 1
        assert 1 in self.flats.keys()

    def tearDown(self) -> None:
        shutil.rmtree(self._folder)
        shutil.rmtree(self._save_dir)

    def testCopy(self):
        """Check behavior of the copy, if mode is manual or automatic"""
        for auto_mode in (True, False):
            with self.subTest(auto_mode=auto_mode):
                if os.path.exists(self._acquisition_without_refs.process_file):
                    os.remove(self._acquisition_without_refs.process_file)

                ref_copy_process = DarkRefsCopy(
                    inputs={
                        "force_sync": True,
                        "data": self._acquisition_without_refs,
                        "auto_mode": auto_mode,
                        "init_dkrf_from": self._acquisition_with_refs,
                        "save_dir": self._save_dir,
                    }
                )
                # make sure the refCopy is correctly initialized
                self.assertTrue(ref_copy_process.has_dark_stored() is True)
                self.assertTrue(ref_copy_process.has_flat_stored())
                # check process is doing the job
                self.assertFalse(
                    os.path.exists(self._acquisition_without_refs.process_file)
                )

                ref_copy_process.run()
                self.assertTrue(
                    os.path.exists(self._acquisition_without_refs.process_file)
                )
                # compare dark and refs
                copied_darks = DarkRefs.get_darks_frm_process_file(
                    self._acquisition_without_refs.process_file,
                    entry=self._acquisition_without_refs.entry,
                )
                self.assertTrue(0 in copied_darks)
                copied_flats = DarkRefs.get_flats_frm_process_file(
                    self._acquisition_without_refs.process_file,
                    entry=self._acquisition_without_refs.entry,
                )
                # when copy dark and flats copy two refs, one at the beginning
                # (index 0) and one at the en (index n_proj)
                self.assertTrue(0 in copied_flats)
                self.assertTrue(9 in copied_flats)

                numpy.testing.assert_allclose(self.darks[0], copied_darks[0])
                numpy.testing.assert_allclose(self.flats[1], copied_flats[0])
                numpy.testing.assert_allclose(self.flats[1], copied_flats[0])



# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__author__ = ["P. Paleo", "H. Payno"]
__license__ = "MIT"
__date__ = "24/05/2016"

from silx.gui import qt
from silx.gui.plot import Plot2D
from . import utils
from tomwer.gui import settings
import numpy
import os
from collections import OrderedDict
import tomwer.resources
from .imagefromfile import ImageFromFile, FileWithImage, _Image
from silx.gui.plot.actions import PlotAction
from silx.gui.plot.actions.control import YAxisInvertedAction
from silx.io.url import DataUrl
import logging
from tomwer.gui import icons
import functools
import time
from collections import namedtuple, Iterable

logger = logging.getLogger(__name__)


class _QImageStackPlot(qt.QWidget):
    """
    Widget to display a stack of image

    :param parent: the Qt parent widget
    """

    _sizeHint = qt.QSize(400, 400)

    IMG_LOADING = numpy.load(
        tomwer.resources._resource_filename(
            "%s.%s" % ("hourglass", "npy"),
            default_directory=os.path.join("gui", "icons"),
        )
    )

    def __init__(self, parent, sliderVertical=False):
        qt.QWidget.__init__(self, parent)
        self._transpose = False
        """If we want to transpose the image like imagj by default"""

        self.images = self.createStackImageInstance()
        layout = qt.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        mainwidget = qt.QWidget(self)
        mainLayout = qt.QHBoxLayout() if sliderVertical else qt.QVBoxLayout()
        mainLayout.setContentsMargins(0, 0, 0, 0)
        mainwidget.setLayout(mainLayout)
        layout.addWidget(mainwidget)

        self._plot = Plot2D(self)

        def find_y_inverted(actions):
            for act in actions:
                if isinstance(act, YAxisInvertedAction):
                    return act
            return None

        y_inverted = find_y_inverted(self._plot.toolBar().actions())
        if y_inverted is None:
            # temporary, XAxisInverted should be embed into silx.
            pos = min(8, len(self._plot.toolBar().actions()))
            y_inverted = self._plot.toolBar().actions()[-pos]
        self.axis_invert = XAxisInvertedAction(plot=self._plot)
        self._plot.toolBar().insertAction(y_inverted, self.axis_invert)

        self._plot.setYAxisInverted(settings.Y_AXIS_DOWNWARD)
        if hasattr(self._plot, "setAxesDisplayed"):
            # by default we want to have a full screen display
            self._plot.setAxesDisplayed(False)

        colormap = self._plot.getDefaultColormap()
        if type(colormap) is dict:
            colormap["autoscale"] = True
        else:
            assert hasattr(colormap, "setVRange")
            colormap.setVRange(None, None)
        self._plot.setDefaultColormap(colormap)
        self._plot.setKeepDataAspectRatio(True)

        # removing some plot action to clear toolbar
        self._plot.getMaskAction().setVisible(False)
        self._plot.getCopyAction().setVisible(False)
        # if has the option medianFilter, active it
        if hasattr(self._plot, "getMedianFilter2DAction"):
            self._plot.getMedianFilter2DAction().setVisible(True)

        mainLayout.addWidget(self._plot)

        if sliderVertical is True:
            lLayout = qt.QVBoxLayout()
        else:
            lLayout = qt.QHBoxLayout()

        self._controlWidget = qt.QWidget(self)
        self._controlWidget.setLayout(lLayout)
        lLayout.setContentsMargins(0, 0, 0, 0)
        mainLayout.addWidget(self._controlWidget)

        self._qspinbox = qt.QSpinBox(parent=self._controlWidget)
        self._qslider = qt.QSlider(
            qt.Qt.Vertical if sliderVertical is True else qt.Qt.Horizontal,
            parent=self._controlWidget,
        )
        self.setRange(0, 0)
        self._qslider.setTickPosition(qt.QSlider.TickPosition(0))
        if sliderVertical:
            lLayout.addWidget(self._qspinbox)
            lLayout.addWidget(self._qslider)
        else:
            lLayout.addWidget(self._qslider)
            lLayout.addWidget(self._qspinbox)
        self._connectControls()
        self.setLayout(layout)

        self.setImages(None)
        self.getActiveImage = self._plot.getActiveImage

    def _disconnectControls(self):
        self._qspinbox.valueChanged.disconnect(self.showImage)
        self._qslider.valueChanged.disconnect(self.showImage)

    def setTranspose(self, value):
        if self._transpose != value:
            self._transpose = value
            img = self._plot.getActiveImage()
            if img is not None:
                data = numpy.transpose(img.getData())
                img.setData(data)

    def _connectControls(self):
        self._qspinbox.valueChanged.connect(self.showImage)
        self._qslider.valueChanged.connect(self.showImage)

    def sizeHint(self):
        """Return a reasonable default size for usage in :class:`PlotWindow`"""
        return self._sizeHint

    def setImages(self, images):
        """
        Set the images to display
        """
        self._qslider.blockSignals(True)
        self._qspinbox.blockSignals(True)

        self.images.clear()
        if images is None:
            self._plot.clear()
            if self._qslider:
                self._qslider.setRange(0, 0)
                self._qspinbox.setRange(0, 0)
            return
        else:
            if isinstance(images, (list, tuple)):
                self.images.setImages(images)
            else:
                assert isinstance(images, (dict, OrderedDict))
                _images = OrderedDict(sorted(images.items(), key=lambda t: t[0]))
                self.images.setImages(_images.values())
            if self.images.size() > 0:
                self.showImage(0)

            if self._qslider:
                if self.images.size() > 0:
                    self.setRange(0, self.images.size())
                else:
                    self.setRange(0, 0)
            if self.images.size() > 0:
                self.showImage(self._qspinbox.value())

        self._qslider.blockSignals(False)
        self._qspinbox.blockSignals(False)

    def createStackImageInstance(self):
        return _StackImage()

    def showImage(self, index):
        """
        Show the image of index 'index' as the current image displayed

        :param int index: the index of the image to display.
        """
        self._qspinbox.blockSignals(True)
        self._qslider.blockSignals(True)
        # update gui
        self._qspinbox.setValue(index)
        self._qslider.setValue(index)
        image = self.images.getImage(index)
        if image is not None:
            if image.data is not None:
                data = image.data
                if self._transpose is True:
                    data = numpy.transpose(data)
                if self.axis_invert.isInverted() is True:
                    data = numpy.fliplr(data)
                self._plot.addImage(data=data, legend=str(index), replace=True)
                self._updateImageInfo()
            else:
                image.load()
        self._qspinbox.blockSignals(False)
        self._qslider.blockSignals(False)

    def setImage(self, index, data):
        """
        Set the given data as the image for the given index

        :param int index: the index of the image to set
        :param numpy.ndarray data: the image
        """
        _data = data
        if data is None:
            _data = self.IMG_NOT_FOUND
        self.images[index] = _data

    def isEmpty(self):
        """

        :return bool: True if no reconstruction has been set yet
        """
        return self.images.empty()

    def clear(self):
        self.setImages(None)

    def setRange(self, _min, _max):
        """Set the range of the spin box

        :param _min: the minimal value of the slice. If equal None, not set
        :param _max: the maximal value of the slice. If equal None, not set
        """
        self._qslider.setRange(_min, _max)
        self._qspinbox.setRange(_min, _max)

    def getControlWidget(self):
        return self._controlWidget

    def _updateImageInfo(self):
        pass


class _QImageFileStackPlot(_QImageStackPlot):
    """
    Widget based on QImageStackPlot but managing images from a stack of path to
    file
    """

    def __init__(self, parent):
        """
        Constructor

        :param parent: the Qt parent widget
        """
        _QImageStackPlot.__init__(self, parent)

        self._loadingMode = _LoadingModeToolButton(parent=self)
        self.getControlWidget().layout().addWidget(self._loadingMode)
        self.addFolderName(False)
        self.clear()

        # connect the qslider with the GUI
        self._sliderConnected = False

        style = qt.QApplication.style()
        self.waitingIcon = style.standardIcon(qt.QStyle.SP_BrowserReload)

        # add some toolbar information
        self._qslider.setToolTip(
            "The loading of the image will start when the \
            slider will be released"
        )
        self._qspinbox.setToolTip(
            "To start the loading press enter on the spin\
             box or release the slider"
        )

        self.layout().addWidget(self.__buildFileInfo())

        self._loadingMode.sigLoadModeChanged.connect(self.setLoadingMode)
        self.setImages(None)

        # expose API
        self.setForceSync = self.images.setForceSync

        # connect signal / slot
        self.images.sigImageLoaded.connect(self._updateIfCurrent)
        self.images.sigLoadingModeChanged.connect(self._loadingModeChanged)
        self._openWithImjButton.released.connect(self._openCurrentInImagej)

    def getLoadingMode(self):
        return self.images.getLoadingMode()

    def setLoadingMode(self, mode):
        self._loadingMode.blockSignals(True)
        self._loadingMode.setLoadingMode(mode)
        self.images.setLoadingMode(mode)
        self.showImage(self._qslider.value())
        self._loadingMode.blockSignals(False)

    def _loadingModeChanged(self, mode, old_mode):
        self._loadingMode.blockSignals(True)
        self._disconnectModeControl(old_mode)
        self._loadingMode.setLoadingMode(mode)
        self._connectModeControl(mode)
        self._loadingMode.blockSignals(False)

    def __buildFileInfo(self):
        self._fileInfoWidget = qt.QWidget(parent=self)
        layout = qt.QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self._fileInfoWidget.setLayout(layout)

        # file name label
        layout.addWidget(qt.QLabel("file :", self._fileInfoWidget), 0, 0, 1, 1)
        self._qlFileName = qt.QLabel("", parent=self._fileInfoWidget)
        layout.addWidget(self._qlFileName, 0, 1, 1, 2)

        # open in image j button
        style = qt.QApplication.style()
        open_icon = style.standardIcon(qt.QStyle.SP_FileLinkIcon)
        self._openWithImjButton = qt.QPushButton(
            open_icon, "open with ImageJ", parent=self
        )
        self._openWithImjButton.setSizePolicy(
            qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum
        )
        layout.addWidget(self._openWithImjButton, 0, 3, 1, 1)

        # date last modification
        layout.addWidget(
            qt.QLabel("last modification :", self._fileInfoWidget), 1, 0, 1, 1
        )
        self._qlLastModifications = qt.QLabel("", parent=self._fileInfoWidget)
        layout.addWidget(self._qlLastModifications, 1, 1, 1, 2)

        return self._fileInfoWidget

    def _disconnectControls(self):
        self._disconnectModeControl(self.getLoadingMode())

    def _connectControls(self):
        self._connectModeControl(self.getLoadingMode())

    def _disconnectModeControl(self, mode):
        if mode == LAZY_LOADING:
            self._qslider.sliderReleased.disconnect(self.changeActiveImageFrmQSlider)
            self._qspinbox.editingFinished.disconnect(self.changeActiveImageFrmQSpinBox)
        else:
            self._qslider.valueChanged.disconnect(self.changeActiveImageFrmQSlider)
            self._qspinbox.valueChanged.disconnect(self.changeActiveImageFrmQSpinBox)

    def _connectModeControl(self, mode):
        if mode == LAZY_LOADING:
            self._qslider.sliderReleased.connect(self.changeActiveImageFrmQSlider)
            self._qspinbox.editingFinished.connect(self.changeActiveImageFrmQSpinBox)
        elif mode in (ON_SHOW_LOADING, ASAP_LOADING):
            self._qslider.valueChanged.connect(self.changeActiveImageFrmQSlider)
            self._qspinbox.valueChanged.connect(self.changeActiveImageFrmQSpinBox)

    def setLoadingImage(self, sliceNb):
        """Set the status of the image to not loaded"""
        self.setImage(sliceNb, _QImageFileStackPlot.IMG_LOADING)

    def setImages(self, imagesFiles):
        """
        Set the stack of images files

        :param _imagesFiles: the list of images path to be displayed
        """
        self._qslider.blockSignals(True)
        self._qspinbox.blockSignals(True)
        _QImageStackPlot.setImages(self, imagesFiles)

        self._qslider.setRange(0, self.images.size() - 1)
        self._qspinbox.setMaximum(self.images.size() - 1)

        self._qslider.blockSignals(False)
        self._qspinbox.blockSignals(False)

    def _openCurrentInImagej(self):
        active_image = self._getActiveImageInfo()
        if active_image is None:
            logger.warning("No active image defined")
        else:
            try:
                utils.open_url_with_image_j(active_image.url)
            except OSError as e:
                msg = qt.QMessageBox(self)
                msg.setIcon(qt.QMessageBox.Warning)
                msg.setWindowTitle("Unable to open image in imagej")
                msg.setText(str(e))
                msg.exec_()

    def _getActiveImageInfo(self):
        imageIndex = self._qslider.value()
        return self.images.getImage(imageIndex)

    def _updateImageInfo(self):
        def slice_to_string(data_slice):
            if data_slice == Ellipsis:
                return "..."
            elif data_slice == slice(None):
                return ":"
            elif isinstance(data_slice, int):
                return str(data_slice)
            else:
                raise TypeError("Unexpected slicing type. Found %s" % type(data_slice))

        image = self._getActiveImageInfo()
        if image is not None:
            imageName = image.url.file_path()
            if self._joinFolderName is True:
                firstFolder = os.path.split(os.path.dirname(imageName))[-1]
                imageName = os.path.join(firstFolder, os.path.basename(imageName))
            else:
                imageName = os.path.basename(imageName)
            if image.url.data_path() is not None:
                imageName = imageName + "&path=" + image.url.data_path()
            if image.url.data_slice() is not None:
                if isinstance(image.url.data_slice(), Iterable):
                    data_slice = [slice_to_string(s) for s in image.url.data_slice()]
                else:
                    data_slice = slice_to_string(image.url.data_slice())
                imageName = imageName + "&slice=" + ",".join(data_slice)

            self._qlFileName.setText(imageName)
            self._qlFileName.setToolTip(image.url.path())
            try:
                last_mod = time.ctime(os.path.getmtime(image.url.file_path()))
            except Exception:
                last_mod = ""
            self._qlLastModifications.setText(last_mod)

    def addFolderName(self, b):
        """

        :param bool val: If true then will join the folder name of the file in
                         addition of the file name.
        """
        self._joinFolderName = b

    def _updateIfCurrent(self, url):
        indexFile = self._qslider.value()
        index = self.images.getUrlIndex(url)
        if index == indexFile:
            self.changeActiveImageFrmQSlider()

    def changeActiveImageFrmQSlider(self):
        indexFile = self._qslider.value()
        self.showImage(indexFile)

    def changeActiveImageFrmQSpinBox(self):
        indexFile = self._qspinbox.value()
        self.showImage(indexFile)

    def getImage(self, index):
        return self._imagesFiles[index]

    def createStackImageInstance(self):
        return _StackImageToLoad()

    def clear(self):
        if hasattr(self, "_qlFileName"):
            self._qlFileName.setText("")
        super(_QImageFileStackPlot, self).clear()


class _StackImage(object):
    def __init__(self, images):
        _images = images
        if _images is None:
            _images = {}
        self._images = OrderedDict(_images)
        """Key: unique index or label or path, Value: _Image"""
        self.nextIndex = 0

    def addImage(self, image):
        assert isinstance(image, _Image)
        self._images[self.nextIndex] = image
        self.nextIndex = self.nextIndex + 1

    def getImage(self, index):
        if index not in self._images:
            return None
        else:
            return self._images[index]

    def size(self):
        return len(self._images)

    def clear(self):
        self._images.clear()
        self.nextIndex = 0


LAZY_LOADING = "lazy loading"
ASAP_LOADING = "load ASAP"
ON_SHOW_LOADING = "load when show requested"

LOADING_MODES = (LAZY_LOADING, ASAP_LOADING, ON_SHOW_LOADING)


class _StackImageToLoad(_StackImage, qt.QObject):
    """
    Define a set of images that can be already loaded or that can be loaded on
    demand
    """

    sigImageLoaded = qt.Signal(str)
    """Signal emitted when an image is loaded. Param is url_path"""
    sigLoadingModeChanged = qt.Signal(str, str)
    """Signal emitted when the loading mode is changed. Parameter are new mode,
    old mode"""

    def __init__(self, loadingMode=LAZY_LOADING):
        assert loadingMode in (LAZY_LOADING, ASAP_LOADING, ON_SHOW_LOADING)
        _StackImage.__init__(self, images=None)
        qt.QObject.__init__(self)
        self._mode = loadingMode
        self._url_to_index = {}
        self._force_sync = False

    def getUrlIndex(self, _url):
        """return the url position on the stack if registred"""
        _url_path = _url
        if isinstance(_url_path, DataUrl):
            _url_path = _url_path.path()
        if _url_path in self._url_to_index:
            return self._url_to_index[_url_path]
        else:
            return None

    def clear(self):
        for img in self._images.values():
            img.sigLoaded.disconnect(self._loaded)
        _StackImage.clear(self)
        self._url_to_index.clear()

    def addImage(self, image):
        if isinstance(image, ImageFromFile):
            image_ = image
        elif isinstance(image, DataUrl):
            image_ = ImageFromFile(url=image)
        else:
            raise TypeError("Unmanaged type:" + type(image))
        self._url_to_index[image_.url.path()] = self.nextIndex
        _StackImage.addImage(self, image_)
        image_.sigLoaded.connect(self._loaded)
        if self._mode == ASAP_LOADING:
            image_.load(self._force_sync)

    def setImages(self, images):
        """

        :param list or tuple images: manage ImageFromFile or file pats (as str)
        """
        self.clear()
        for img in images:
            if type(img) is str:
                self.addFile(_file=img)
            else:
                self.addImage(img)

    def setFiles(self, files):
        self.clear()
        for _file in files:
            self.addFile(_file=_file)

    def addFile(self, _file):
        _fileWithImage = FileWithImage(_file)
        _imagesFiles = _fileWithImage.getImages(_load=self._mode == ASAP_LOADING)
        if _imagesFiles is None:
            return
        for imgFile in _imagesFiles:
            self.addImage(imgFile)

    def setLoadingMode(self, mode):
        assert mode in LOADING_MODES
        old = self._mode
        self._mode = mode
        self.sigLoadingModeChanged.emit(mode, old)
        if self._mode == ASAP_LOADING:
            self._loadAllImages()

    def getLoadingMode(self):
        return self._mode

    def _loadAllImages(self):
        for imgIndex, img in self._images.items():
            img.load_recons_params(self._force_sync)

    def _loaded(self, url_path):
        self.sigImageLoaded.emit(url_path)

    def setForceSync(self, value):
        self._force_sync = value


_loadingModeState = namedtuple("_loadingModeState", ["icon", "state", "action"])


class _LoadingModeToolButton(qt.QToolButton):
    """Tool button to switch keep aspect ratio of a plot"""

    sigLoadModeChanged = qt.Signal(str)

    def __init__(self, parent=None):
        super(_LoadingModeToolButton, self).__init__(parent=parent)

        self._states = {
            ON_SHOW_LOADING: _loadingModeState(
                icon=icons.getQIcon("low_speed"),
                state="load data only when requiring",
                action="will load data only when requiring",
            ),
            LAZY_LOADING: _loadingModeState(
                icon=icons.getQIcon("medium_low_speed"),
                state="load data when seems convenient",
                action="will load data when seems convenient",
            ),
            ASAP_LOADING: _loadingModeState(
                icon=icons.getQIcon("high_speed"),
                state="load data as soon as possible",
                action="will load data as soon as possible",
            ),
        }

        onShowLoadingMode = self._createAction(ON_SHOW_LOADING)
        fctOSL = functools.partial(self.setLoadingMode, ON_SHOW_LOADING)
        onShowLoadingMode.triggered.connect(fctOSL)

        lazyLoadingMode = self._createAction(LAZY_LOADING)
        fctLL = functools.partial(self.setLoadingMode, LAZY_LOADING)
        lazyLoadingMode.triggered.connect(fctLL)

        asapLoadingMode = self._createAction(ASAP_LOADING)
        fctAL = functools.partial(self.setLoadingMode, ASAP_LOADING)
        asapLoadingMode.triggered.connect(fctAL)

        menu = qt.QMenu(self)
        menu.addAction(onShowLoadingMode)
        menu.addAction(lazyLoadingMode)
        menu.addAction(asapLoadingMode)
        self.setMenu(menu)
        self.setPopupMode(qt.QToolButton.InstantPopup)

        self.setLoadingMode(ON_SHOW_LOADING)

    def _createAction(self, loadingMode):
        icon = self._states[loadingMode].icon
        text = self._states[loadingMode].action
        return qt.QAction(icon, text, self)

    def setLoadingMode(self, loadingMode):
        assert loadingMode in LOADING_MODES
        icon = self._states[loadingMode].icon
        toolTip = self._states[loadingMode].state
        self.setIcon(icon)
        self.setToolTip(toolTip)
        self.sigLoadModeChanged.emit(loadingMode)


class XAxisInvertedAction(PlotAction):
    """QAction controlling X orientation on a :class:`.PlotWidget`.

    :param plot: :class:`.PlotWidget` instance on which to operate
    :param parent: See :class:`QAction`
    """

    def __init__(self, plot, parent=None):
        # Uses two images for checked/unchecked states
        self._states = {
            False: (icons.getQIcon("plot-xleft"), "Orient X axis left"),
            True: (icons.getQIcon("plot-xright"), "Orient X axis right"),
        }

        self._inverted = False
        icon, tooltip = self._states[self.isInverted()]
        super(XAxisInvertedAction, self).__init__(
            plot,
            icon=icon,
            text="Invert X Axis",
            tooltip=tooltip,
            triggered=self._actionTriggered,
            checkable=False,
            parent=parent,
        )

    def _xAxisInvertedChanged(self, inverted):
        """Handle Plot set y axis inverted signal"""
        icon, tooltip = self._states[inverted]
        self.setIcon(icon)
        self.setToolTip(tooltip)

    def _actionTriggered(self, checked=False):
        self._inverted = not self._inverted
        icon, tooltip = self._states[self.isInverted()]
        # this only work for images
        self.setIcon(icon)
        self.setToolTip(tooltip)

        # flip the current data
        data = self.plot.getActiveImage()
        if data:
            data_array = numpy.fliplr(data.getData())
            data.setData(data_array)

    def isInverted(self):
        return self._inverted



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "05/03/2019"


from .base import _ReconsParam
from ....utils import _assert_param_instance, _assert_cast_to_boolean
from silx.utils.enum import Enum as _Enum
from collections.abc import Iterable


class PaganinMode(_Enum):
    off = 0
    on = 1
    both = 2
    multi = 3


class PaganinRP(_ReconsParam):
    """
    Paganin parameters to use during ftseries process
    """

    def __init__(self):
        _ReconsParam.__init__(self)
        self.__mode = PaganinMode.off
        self.__db = 500.0
        self.__db2 = 100.0
        self.__unsharp_sigma = 0.8
        self.__unsharp_coeff = 3.0
        self.__threshold = 500
        self.__dilate = 2
        self.__median_r = 4
        self.__keep_bone = False
        self.__keep_soft = False
        self.__keep_abs = False
        self.__keep_corr = False
        self.__keep_mask = False

        self._managed_params = {
            "MODE": self.__class__.mode,
            "DB": self.__class__.db,
            "DB2": self.__class__.db2,
            "UNSHARP_SIGMA": self.__class__.unsharp_sigma,
            "UNSHARP_COEFF": self.__class__.unsharp_coeff,
            "THRESHOLD": self.__class__.threshold,
            "DILATE": self.__class__.dilate,
            "MEDIANR": self.__class__.median_r,
            "MKEEP_BONE": self.__class__.mkeep_bone,
            "MKEEP_SOFT": self.__class__.mkeep_soft,
            "MKEEP_ABS": self.__class__.mkeep_abs,
            "MKEEP_CORR": self.__class__.mkeep_corr,
            "MKEEP_MASK": self.__class__.mkeep_mask,
        }

    @property
    def mode(self):
        return self.__mode

    @mode.setter
    def mode(self, mode):
        _assert_param_instance(mode, (PaganinMode, int))
        _mode = mode
        if isinstance(_mode, int):
            _mode = PaganinMode(_mode)
        self.__mode = _mode
        self.changed()

    @property
    def db(self):
        """value of delta/beta"""
        return self.__db

    @staticmethod
    def _get_db_fromstr(vals):
        vals = vals.replace(" ", "")
        vals = vals.replace("(", "")
        vals = vals.replace(")", "")
        vals = vals.replace("]", "")
        vals = vals.replace("[", "")
        vals = vals.replace(";", ",").split(",")
        if vals[-1] == "":
            del vals[-1]

        res = []
        [res.append(float(val)) for val in vals]
        if len(res) == 1:
            return res[0]
        else:
            return tuple(res)

    @db.setter
    def db(self, value):
        _assert_param_instance(value, (int, float, str))
        if isinstance(value, str):
            _db = self._get_db_fromstr(vals=value)
        else:
            _db = value
        if _db != self.__db:
            self.__db = _db
            self.changed()

    @property
    def db2(self):
        return self.__db2

    @db2.setter
    def db2(self, value):
        _assert_param_instance(value, (int, float, str))
        if isinstance(value, str):
            _db2 = self._get_db_fromstr(vals=value)
        else:
            _db2 = value
        if _db2 != self.__db2:
            self.__db2 = _db2
            self.changed()

    @property
    def unsharp_sigma(self):
        """size of the mask of unsharp masking"""
        return self.__unsharp_sigma

    @unsharp_sigma.setter
    def unsharp_sigma(self, value):
        """coeff for unsharp masking"""
        _assert_param_instance(value, float)
        if self.__unsharp_sigma != value:
            self.__unsharp_sigma = value
            self.changed()

    @property
    def unsharp_coeff(self):
        return self.__unsharp_coeff

    @unsharp_coeff.setter
    def unsharp_coeff(self, value):
        _assert_param_instance(value, float)
        if self.__unsharp_coeff != value:
            self.__unsharp_coeff = value
            self.changed()

    @property
    def threshold(self):
        return self.__threshold

    @threshold.setter
    def threshold(self, value):
        _assert_param_instance(value, (int, float))
        if self.__threshold != float(value):
            self.__threshold = float(value)
            self.changed()

    @property
    def dilate(self):
        return self.__dilate

    @dilate.setter
    def dilate(self, value):
        if self.__dilate != value:
            self.__dilate = value
            self.changed()

    @property
    def median_r(self):
        return self.__median_r

    @median_r.setter
    def median_r(self, value):
        if self.__median_r != value:
            self.__median_r = value
            self.changed()

    @property
    def mkeep_bone(self):
        return self.__keep_bone

    @mkeep_bone.setter
    def mkeep_bone(self, keep):
        # TODO: float should be removed, but this is a legacy from fastomo3
        _assert_param_instance(keep, (int, bool, float))
        _assert_cast_to_boolean(keep)
        if self.__keep_bone != bool(keep):
            self.__keep_bone = bool(keep)
            self.changed()

    @property
    def mkeep_soft(self):
        return self.__keep_soft

    @mkeep_soft.setter
    def mkeep_soft(self, keep):
        # TODO: float should be removed, but this is a legacy from fastomo3
        _assert_param_instance(keep, (int, bool, float))
        _assert_cast_to_boolean(keep)
        if self.__keep_soft != bool(keep):
            self.__keep_soft = bool(keep)
            self.changed()

    @property
    def mkeep_abs(self):
        return self.__keep_abs

    @mkeep_abs.setter
    def mkeep_abs(self, keep):
        # TODO: float should be removed, but this is a legacy from fastomo3
        _assert_param_instance(keep, (int, bool, float))
        _assert_cast_to_boolean(keep)
        if self.__keep_abs != bool(keep):
            self.__keep_abs = bool(keep)
            self.changed()

    @property
    def mkeep_corr(self):
        return self.__keep_corr

    @mkeep_corr.setter
    def mkeep_corr(self, keep):
        _assert_param_instance(keep, (int, bool, float))
        _assert_cast_to_boolean(keep)
        if self.__keep_corr != bool(keep):
            self.__keep_corr = bool(keep)
            self.changed()

    @property
    def mkeep_mask(self):
        return self.__keep_mask

    @mkeep_mask.setter
    def mkeep_mask(self, keep):
        _assert_param_instance(keep, (int, bool, float))
        _assert_cast_to_boolean(keep)
        if self.__keep_mask != bool(keep):
            self.__keep_mask = bool(keep)
            self.changed()

    def to_dict(self):
        _dict = {
            "MODE": self.mode.value,
            "DB": str(self.db),
            "DB2": str(self.db2),
            "UNSHARP_SIGMA": self.unsharp_sigma,
            "UNSHARP_COEFF": self.unsharp_coeff,
            "THRESHOLD": self.threshold,
            "DILATE": self.dilate,
            "MEDIANR": self.median_r,
            "MKEEP_BONE": self.mkeep_bone,
            "MKEEP_SOFT": self.mkeep_soft,
            "MKEEP_ABS": self.mkeep_abs,
            "MKEEP_CORR": self.mkeep_corr,
            "MKEEP_MASK": self.mkeep_mask,
        }
        _dict.update(self.unmanaged_params)
        return _dict

    @staticmethod
    def from_dict(_dict: dict):
        recons_param = PaganinRP()
        recons_param.load_from_dict(_dict)
        return recons_param

    def load_from_dict(self, _dict: dict) -> None:
        self._load_unmanaged_params(_dict)
        self.mode = PaganinMode(_dict["MODE"])
        self.db = _dict["DB"]
        self.db2 = _dict["DB2"]
        self.unsharp_sigma = _dict["UNSHARP_SIGMA"]
        self.unsharp_coeff = _dict["UNSHARP_COEFF"]
        self.threshold = _dict["THRESHOLD"]
        self.dilate = _dict["DILATE"]
        self.median_r = _dict["MEDIANR"]
        self.mkeep_bone = _dict["MKEEP_BONE"]
        self.mkeep_soft = _dict["MKEEP_SOFT"]
        self.mkeep_abs = _dict["MKEEP_ABS"]
        self.mkeep_corr = _dict["MKEEP_CORR"]
        self.mkeep_mask = _dict["MKEEP_MASK"]

    def has_several_db_param(self) -> bool:
        """

        :return: True if the reconstruction contains several delta/beta values.
                 If the paganin mode is off will return False by default,
                 No matter the values of db and db2
        """
        if self.mode is PaganinMode.off:
            return False
        if isinstance(self.db, Iterable) and len(self.db) > 1:
            return True
        if isinstance(self.db2, Iterable) and len(self.db2) > 1:
            return True
        return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import sys
import os
from silx.gui import qt
import argparse
from tomwer.gui.utils.splashscreen import getMainSplashScreen
from tomwer.gui.stacks import RadioStack
from tomwer.gui import icons
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.core.utils.resource import increase_max_number_file
import signal

logging.basicConfig(level=logging.WARNING)
_logger = logging.getLogger(__name__)


def getinputinfo():
    return "tomwer radiostack [scanDir]"


def addFolderAndSubFolder(stack, path):
    stack.addLeafScan(path)
    for f in os.listdir(path):
        _path = os.path.join(path, f)
        if os.path.isdir(_path) is True:
            addFolderAndSubFolder(stack, _path)
        elif os.path.isfile(_path) and HDF5TomoScan.is_nexus_nxtomo_file(_path):
            stack.addLeafScan(_path)


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    qt.QApplication.quit()


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "root_dir", help="Root dir to browse and to extract all radio under it."
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Set logging system in debug mode",
    )

    increase_max_number_file()

    options = parser.parse_args(argv[1:])

    global app  # QApplication must be global to avoid seg fault on quit
    app = qt.QApplication.instance() or qt.QApplication([])

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler
    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # cached
    timer.timeout.connect(lambda: None)

    splash = getMainSplashScreen()
    widget = RadioStack()
    widget.setWindowIcon(icons.getQIcon("tomwer"))
    options.root_dir = options.root_dir.rstrip(os.path.sep)
    addFolderAndSubFolder(stack=widget, path=options.root_dir)

    splash.finish(widget)
    widget.show()
    app.exec_()


if __name__ == "__main__":
    main(sys.argv)



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/
"""
Utils to mock scans
"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "09/01/2020"


import shutil
import glob
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from silx.utils.enum import Enum as _Enum
from silx.io.url import DataUrl
import os
from tomwer.synctools.rsyncmanager import RSyncManager


class ScanStages:
    """
    Util class to copy all the files of scan to a destination dir until
    a define advancement of the acquisition.

    :param TomwerScanBase scan: scan to copy.
    """

    class AcquisitionStage(_Enum):
        ACQUI_NOT_STARTED = (0,)
        ACQUI_STARTED = (1,)
        ACQUI_ON_GOING = (2,)
        ACQUI_ENDED = (3,)
        RECONSTRUCTION_ADDED = (4,)
        COMPLETE = (99,)

        def get_command_name(self):
            """Return the name of the AcquisitionStage to give by a command
            option"""
            return self.name.lower().replace("_", "-")

        @staticmethod
        def get_command_names():
            return [stage.get_command_name() for stage in ScanStages.AcquisitionStage]

        @staticmethod
        def from_command_name(name):
            """Return the AcquisitionStage fitting a command option"""
            name_ = name.replace("-", "_").upper()
            return getattr(ScanStages.AcquisitionStage, name_)

    def __init__(self, scan: TomwerScanBase):
        assert isinstance(scan, TomwerScanBase)
        self.scan = scan

    def rsync_until(self, stage: AcquisitionStage, dest_dir: str) -> None:
        """

        :param stage:
        :type: AcquisitionStage
        :param dest_dir:
        :type: str
        """
        stage = ScanStages.AcquisitionStage.from_value(stage)
        if not dest_dir.endswith(os.path.basename(self.scan.path)):
            dest_dir = os.path.join(dest_dir, os.path.basename(self.scan.path))
        for t_stage in ScanStages.AcquisitionStage:
            if t_stage.value <= stage.value:
                self._rsync_stage(t_stage, dest_dir=dest_dir)

    def _rsync_stage(self, stage: AcquisitionStage, dest_dir: str) -> None:
        """
        Synchronize a specific stage of scan to the given destination directory

        :param stage: stage to copy
        :type: AcquisitionStage
        :param dest_dir:
        :type: str
        """
        if not dest_dir.endswith(os.path.basename(self.scan.path)):
            dest_dir = os.path.join(dest_dir, os.path.basename(self.scan.path))
        stage = ScanStages.AcquisitionStage.from_value(stage)
        if stage is ScanStages.AcquisitionStage.ACQUI_NOT_STARTED:
            return
        elif stage is ScanStages.AcquisitionStage.ACQUI_STARTED:
            if isinstance(self.scan, EDFTomoScan):
                if not os.path.isdir(dest_dir):
                    os.makedirs(dest_dir)
                info_file = self.scan.get_info_file(self.scan.path)
                if info_file is not None:
                    file_info_dest = os.path.join(dest_dir, os.path.basename(info_file))
                    RSyncManager().sync_file(
                        source=info_file, target=file_info_dest, wait=True
                    )
            elif isinstance(self.scan, HDF5TomoScan):
                raise NotImplementedError()
        elif stage is ScanStages.AcquisitionStage.ACQUI_ON_GOING:
            # copy half of the acquisition file
            if isinstance(self.scan, EDFTomoScan):
                urls = self.scan.projections
                n_url = len(urls) // 2
                keys = list(urls.keys())
                while n_url > 0:
                    n_url -= 1
                    self._copy_url_file(url=urls[keys[n_url]], dest_dir=dest_dir)

            elif isinstance(self.scan, HDF5TomoScan):
                raise NotImplementedError()
        elif stage is ScanStages.AcquisitionStage.ACQUI_ENDED:
            if isinstance(self.scan, EDFTomoScan):
                for _, url in self.scan.projections.items():
                    file_target = url.file_path().replace(self.scan.path, dest_dir)
                    if not os.path.exists(file_target):
                        self._copy_url_file(url=url, dest_dir=dest_dir)
                xml_file = os.path.join(
                    self.scan.path, os.path.basename(self.scan.path) + ".xml"
                )
                xml_dest = os.path.join(dest_dir, os.path.basename(xml_file))
                RSyncManager().sync_file(source=xml_file, target=xml_dest, wait=True)
            elif isinstance(self.scan, HDF5TomoScan):
                raise NotImplementedError()
        elif stage is ScanStages.AcquisitionStage.RECONSTRUCTION_ADDED:
            # copy reconstruction
            if isinstance(self.scan, EDFTomoScan):
                pyhst_files = EDFTomoScan.get_pyhst_recons_file(scanID=self.scan.path)
                if pyhst_files:
                    for par_file in pyhst_files:
                        par_file_dst = os.path.join(
                            dest_dir, os.path.basename(par_file)
                        )
                        RSyncManager().sync_file(
                            source=par_file, target=par_file_dst, wait=True
                        )
                for reconstructed_file in EDFTomoScan.get_reconstructions_paths(
                    scanID=self.scan.path
                ):
                    assert os.path.isfile(reconstructed_file)
                    recons_file_dest = os.path.join(
                        dest_dir, os.path.basename(reconstructed_file)
                    )
                    RSyncManager().sync_file(
                        source=reconstructed_file, target=recons_file_dest, wait=True
                    )

                for par_file in glob.glob(os.path.join(self.scan.path, "*.par")):
                    par_file_src = os.path.join(self.scan.path, par_file)
                    par_file_dst = os.path.join(dest_dir, os.path.basename(par_file))
                    RSyncManager().sync_file(
                        source=par_file_src, target=par_file_dst, wait=True
                    )
                # copy some info file that can came from .vol files
                for info_file in glob.glob(os.path.join(self.scan.path, "*.info")):
                    info_file_src = os.path.join(self.scan.path, info_file)
                    info_file_dst = os.path.join(dest_dir, os.path.basename(info_file))
                    RSyncManager().sync_file(
                        source=info_file_src, target=info_file_dst, wait=True
                    )
                # copy some xml file that can came from pyhst
                xml_files = glob.glob(os.path.join(self.scan.path, "*.xml"))
                for xml_file in xml_files:
                    _xml_file = os.path.join(self.scan.path, xml_file)
                    xml_dest = os.path.join(dest_dir, os.path.basename(_xml_file))
                    RSyncManager().sync_file(
                        source=_xml_file, target=xml_dest, wait=True
                    )
            elif isinstance(self.scan, HDF5TomoScan):
                raise NotImplementedError()
        elif stage is ScanStages.AcquisitionStage.COMPLETE:
            if isinstance(self.scan, EDFTomoScan):
                for file_ in os.listdir(self.scan.path):
                    file_fp = os.path.join(self.scan.path, file_)
                    RSyncManager().sync_file(source=file_fp, target=dest_dir, wait=True)
            elif isinstance(self.scan, HDF5TomoScan):
                raise NotImplementedError()
        else:
            raise ValueError("given stage is not recognized", stage)

    def _copy_url_file(self, url, dest_dir):
        assert isinstance(url, DataUrl)
        file_target = url.file_path().replace(self.scan.path, dest_dir)
        if not os.path.exists(file_target):
            shutil.copyfile(src=url.file_path(), dst=file_target)



# /*##########################################################################
# Copyright (C) 2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/
"""Some widget construction to check if a sample moved"""

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "19/03/2018"


from collections import OrderedDict
from tomwer.gui.samplemoved.selectiontable import AngleSelectionTable
from silx.gui import qt
import silx.io.utils
import silx.io.url
import logging

_logger = logging.getLogger(__name__)
try:
    from silx.gui.plot.CompareImages import CompareImages
except ImportError:
    _logger.warning("silx >0.7 should be installed to access the SampleMovedWidget")


class SampleMovedWidget(qt.QMainWindow):
    """
    Widget used to display two images with different color channel.
    The goal is to see if the sample has moved during acquisition.
    """

    CONFIGURATIONS = OrderedDict(
        [
            ("0-0(1)", (("0", "0.0", 0), ("0(1)", "0.0 (1)"))),
            ("90-90(1)", (("90", "90.0", 90), ("90(1)", "90.0 (1)"))),
            ("180-180(1)", (("180", "180.0", 180), ("180(1)", "180.0 (1)"))),
            ("270-270(1)", (("270", "270.0", 270), ("270(1)", "270.0 (1)"))),
            ("360-0", (("360", "360.0", 360), ("0", "0.0"))),
        ]
    )
    """Define possible configurations for comparison. Key is the name of the
    configuration, value contains a couple valid values for the necessary
    two projections
    """

    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)
        self.setWindowFlags(qt.Qt.Widget)
        self._isConnected = False
        self._images = {}
        self._symmetricalStates = {"first": False, "second": False}
        self._plot = CompareImages(parent=self)
        self._plot.setWindowFlags(qt.Qt.Widget)
        self._on_load_callback = []

        self._topWidget = self.getControlWidget()

        self._dockWidgetMenu = qt.QDockWidget(parent=self)
        self._dockWidgetMenu.layout().setContentsMargins(0, 0, 0, 0)
        self._dockWidgetMenu.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self._dockWidgetMenu.setWidget(self._topWidget)
        self.addDockWidget(qt.Qt.BottomDockWidgetArea, self._dockWidgetMenu)

        self._plotsWidget = qt.QWidget(parent=self)
        self._plotsWidget.setLayout(qt.QHBoxLayout())

        self._plotsWidget.layout().addWidget(self._plot)
        self.setCentralWidget(self._plotsWidget)

        if hasattr(self._selectorCB, "currentTextChanged"):
            self._selectorCB.currentTextChanged.connect(self.setConfiguration)
        else:
            self._selectorCB.currentIndexChanged["QString"].connect(
                self.setConfiguration
            )

        self._selectionTable.sigImageAChanged.connect(self._setConfigManual)
        self._selectionTable.sigImageBChanged.connect(self._setConfigManual)

        # expose API
        self.setSelection = self._selectionTable.setSelection

    def getControlWidget(self):
        if hasattr(self, "_topWidget"):
            return self._topWidget
        self._topWidget = qt.QWidget(parent=self)

        self._configWidget = qt.QWidget(parent=self._topWidget)
        self._configWidget.setLayout(qt.QHBoxLayout())

        self._configWidget.layout().addWidget(
            qt.QLabel("Configuration:", parent=self._topWidget)
        )
        self._selectorCB = qt.QComboBox(parent=self._topWidget)
        self._configWidget.layout().addWidget(self._selectorCB)

        self._selectionTable = AngleSelectionTable(parent=self._topWidget)
        self._topWidget.setLayout(qt.QVBoxLayout())
        self._topWidget.layout().setContentsMargins(0, 0, 0, 0)

        self._topWidget.layout().addWidget(self._configWidget)
        self._topWidget.layout().addWidget(self._selectionTable)

        self._selectionTable.sigImageAChanged.connect(self._changeImageA)
        self._selectionTable.sigImageBChanged.connect(self._changeImageB)
        return self._topWidget

    def setOnLoadAction(self, action):
        self._on_load_callback.append(action)

    def clearOnLoadActions(self):
        self._on_load_callback = []

    def clear(self):
        self._selectorCB.clear()
        self._selectionTable.clear()
        self._images = {}

    def setImages(self, images):
        """
        Set the images in a key value system. Key should be in
        (0, 90, 180, 270) and the value should be the image.

        images value can be str (path to the file) or data

        :param dict images: images to set. key is index or file name, value
                            the image.
        """
        self.clear()
        self._images = images

        # update the default config
        self._selectorCB.clear()

        def contains_at_least_one_key(keys):
            for key in keys:
                if key in images.keys():
                    return True
            return False

        self._selectorCB.blockSignals(True)
        for config in self.CONFIGURATIONS:
            proj_0_keys, proj_1_keys = self.CONFIGURATIONS[config]
            if contains_at_least_one_key(proj_0_keys) and contains_at_least_one_key(
                proj_1_keys
            ):
                self._selectorCB.addItem(config)
        self._selectorCB.addItem("manual")

        for angleValue, file_path in images.items():
            self._selectionTable.addRadio(name=file_path, angle=angleValue)
        self._selectorCB.setCurrentIndex(0)
        self._selectorCB.blockSignals(False)
        if hasattr(self._selectorCB, "currentTextChanged"):
            self._selectorCB.currentTextChanged.emit(self._selectorCB.currentText())
        else:
            self._selectorCB.currentIndexChanged["QString"].emit(
                self._selectorCB.currentText()
            )

    def _updatePlot(self):
        imgA, imgB = self._selectionTable.getSelection()
        dataImgA = dataImgB = None
        if imgA is not None:
            dataImgA = silx.io.utils.get_data(silx.io.url.DataUrl(path=imgA))
            for callback in self._on_load_callback:
                callback(dataImgA)
        if imgB is not None:
            dataImgB = silx.io.utils.get_data(silx.io.url.DataUrl(path=imgB))
            for callback in self._on_load_callback:
                callback(dataImgB)
        if dataImgA is not None and dataImgB is not None:
            self._plot.setData(image1=dataImgA, image2=dataImgB)

    def _changeImageA(self, img):
        if img is not None:
            dataImgA = silx.io.utils.get_data(silx.io.url.DataUrl(path=img))
            for callback in self._on_load_callback:
                callback(dataImgA)
            self._plot.setImage1(image1=dataImgA)

    def _changeImageB(self, img):
        if img is not None:
            dataImgB = silx.io.utils.get_data(silx.io.url.DataUrl(path=img))
            for callback in self._on_load_callback:
                callback(dataImgB)
            self._plot.setImage2(image2=dataImgB)

    def setConfiguration(self, config):
        if config == "manual":
            return
        if config not in self.CONFIGURATIONS:
            _logger.warning("Undefined configuration: %s" % config)
            return

        self._selectionTable.blockSignals(True)
        self._selectionTable.setAngleSelection(
            self.CONFIGURATIONS[config][0], self.CONFIGURATIONS[config][1]
        )
        self._updatePlot()
        self._selectionTable.blockSignals(False)

    def _setConfigManual(self):
        indexItemManual = self._selectorCB.findText("manual")
        if indexItemManual >= 0:
            self._selectorCB.setCurrentIndex(indexItemManual)



# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "10/01/2018"


import logging
from silx.gui import qt
from tomwer.core.utils import pyhstutils
from tomwer.gui.reconstruction.ftserie.h5editor import H5StructEditor
from tomwer.synctools.ftseries import QReconsParams, _QPyhstRP
from tomwer.gui.reconstruction.deviceselector import CudaPlatfornGroup

logger = logging.getLogger(__name__)


class PyHSTWidget(H5StructEditor, qt.QWidget):
    """
    Definition of the PyHST tab to edit the PyHST parameters

    :param reconsparams: reconstruction parameters edited by the widget
    """

    def __init__(self, reconsparams, parent=None):
        qt.QWidget.__init__(self, parent)
        H5StructEditor.__init__(self, structID="PYHSTEXE")
        self._recons_params = None
        self.setReconsParams(recons_params=reconsparams)

        self.setLayout(qt.QVBoxLayout())
        self.layout().addWidget(self.__buildPYHSTVersion())
        self.layout().addWidget(self.__buildPYHSTOfficialVersion())

        self._qcbverbose = qt.QCheckBox("verbose", parent=self)
        self.layout().addWidget(self._qcbverbose)
        self.linkCheckboxWithH5Variable(self._qcbverbose, "VERBOSE", invert=False)

        self.layout().addWidget(self.__buildVerboseFile())
        self.layout().addWidget(self.__buildMakeOAR())
        self.layout().addWidget(self.__buildCudaOptions())

        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer)

        self._qcbverbose.setChecked(False)
        self._versoseFileWidget.setVisible(True)
        self._qcbverbose.toggled.connect(self._versoseFileWidget.setDisabled)

        self._makeConnection()
        # expose API
        self.getCudaDevices = self._cudaSelector.getExistingDevices

    def setReconsParams(self, recons_params):
        if isinstance(recons_params, QReconsParams):
            _recons_params = recons_params.pyhst
        elif isinstance(recons_params, _QPyhstRP):
            _recons_params = recons_params
        else:
            raise ValueError(
                "recons_params should be an instance of QReconsParam or _QPyhstRP"
            )

        if self._recons_params:
            self._recons_params.sigChanged.disconnect(self._update_params)
        self._recons_params = _recons_params
        self.load(self._recons_params)
        self._recons_params.sigChanged.connect(self._update_params)

    def _update_params(self):
        """Update all parameter"""
        self.load(self._recons_params)

    def _makeConnection(self):
        self._qcbverbose.toggled.connect(self._verboseChanged)
        self._qcbPyHSTVersion.currentIndexChanged.connect(self._pyHSTVersionChanged)
        self._qleVerboseFile.editingFinished.connect(self._verboseFileChanged)
        self._makeOARFileCB.toggled.connect(self._makeOARChanged)
        self._cudaSelector.selectionChanged.connect(self._cudaDevicesChanged)

    def _verboseChanged(self, b):
        self._recons_params["VERBOSE"] = b

    def __buildPYHSTVersion(self):
        widget = qt.QWidget(self)
        widget.setLayout(qt.QHBoxLayout())

        # create the PyHST version combobox
        widget.layout().addWidget(qt.QLabel("PyHST version : ", parent=widget))
        self._qcbPyHSTVersion = qt.QComboBox(parent=self)
        widget.layout().addWidget(self._qcbPyHSTVersion)
        self.linkComboboxWithH5Variable(
            self._qcbPyHSTVersion, "EXE", fitwithindex=False, setDefault=False
        )

        # try to get the PyHST dir
        d = pyhstutils._getPyHSTDir()
        if d is None:
            raise logger.warning(
                """Can't find the directory containing the PyHST
                directory. Please set the environment variable
                PYHST_DIR and run again"""
            )

        availablePyHSTVersion = pyhstutils._findPyHSTVersions(d)
        availablePyHSTVersion.extend(pyhstutils._getPyHSTFromEnvVariable())
        if len(availablePyHSTVersion) == 0:
            self.__warmNoPyHSTFound(d)
            pass
        else:
            [self._qcbPyHSTVersion.addItem(exe) for exe in availablePyHSTVersion]

        return widget

    def __buildMakeOAR(self):
        self._makeOARFileCB = qt.QCheckBox("make OAR file", parent=self)
        self.linkCheckboxWithH5Variable(
            qcheckbox=self._makeOARFileCB, h5ParamName="MAKE_OAR_FILE"
        )
        self._makeOARFileCB.toggled.connect(self._makeOARChanged)
        return self._makeOARFileCB

    def __buildCudaOptions(self):
        self._cudaSelector = CudaPlatfornGroup(parent=self)
        # by default activate all cuda device
        self._cudaSelector.activate_all()
        self.linkGroupWithH5Variable(
            group=self._cudaSelector,
            h5ParamName="CUDA_DEVICES",
            setter=self._cudaSelector.setDevices,
            getter=self._cudaSelector.getSelectedDevices,
        )
        old = self._recons_params.blockSignals(True)
        self._recons_params["CUDA_DEVICES"] = self._cudaSelector.getSelectedDevices()
        self._recons_params.blockSignals(old)

        return self._cudaSelector

    def _pyHSTVersionChanged(self):
        value = self._qcbPyHSTVersion.currentText()
        self._recons_params["EXE"] = value

    def _makeOARChanged(self, b):
        self._recons_params["MAKE_OAR_FILE"] = b

    def _cudaDevicesChanged(self, devices):
        self._recons_params.cuda_devices = devices

    def __buildPYHSTOfficialVersion(self):
        """build the official version QLine edit and update the _qcbPyHSTVersion
        combobox so should always be called after.
        """
        widget = qt.QWidget(self)
        widget.setLayout(qt.QHBoxLayout())
        widget.layout().addWidget(qt.QLabel("PyHST official version", parent=widget))
        self._qlOfficalVersion = qt.QLabel("", parent=widget)
        widget.layout().addWidget(self._qlOfficalVersion)
        self.linkGroupWithH5Variable(
            self._qlOfficalVersion,
            "OFFV",
            getter=self._qlOfficalVersion.text,
            setter=self._qlOfficalVersion.setText,
        )
        return widget

    def __warmNoPyHSTFound(self, directory):
        """Simple function displaying a MessageBox that PyHST haven't been found"""
        text = "No executable of PyHST have been found in %s." % directory
        text += " You might set the environment variable PYHST_DIR "
        text += " or install PyHST."
        logger.info(text)

    def __buildVerboseFile(self):
        self._versoseFileWidget = qt.QWidget(self)
        self._versoseFileWidget.setLayout(qt.QHBoxLayout())
        self._versoseFileWidget.layout().addWidget(
            qt.QLabel("name of the PyHST information output file", parent=self)
        )
        self._qleVerboseFile = qt.QLineEdit("", parent=None)
        self._versoseFileWidget.layout().addWidget(self._qleVerboseFile)
        self.LinkLineEditWithH5Variable(self._qleVerboseFile, "VERBOSE_FILE")
        return self._versoseFileWidget

    def _verboseFileChanged(self):
        value = self._qleVerboseFile.text()
        self._recons_params._set_parameter_value(parameter="VERBOSE_FILE", value=value)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "14/02/2017"


import logging
import pytest
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt
from tomwer.core.process.reconstruction.ftseries.params.beamgeo import BeamGeoType
from tomwer.core.process.reconstruction.darkref.params import Method as dkrf_method
from tomwer.core.process.reconstruction.ftseries.params.ft import VolSelMode
from tomwer.core.process.reconstruction.ftseries.params.paganin import PaganinMode
from tomwer.core.process.reconstruction.ftseries.params.pyhst import PyhstRP
from tomwer.gui.reconstruction.ftserie.reconsparamseditor import ReconsParamsEditor
from tomwer.synctools.ftseries import QReconsParams
from tomwer.test.utils import skip_gui_test
import gc

logging.disable(logging.INFO)


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestSimplifyH5EditorDisplay(TestCaseQt):
    """Make sure we are displayig the correct things"""

    def setUp(self):
        super(TestSimplifyH5EditorDisplay, self).setUp()
        self.h5Editor = ReconsParamsEditor(None)
        self.h5Editor.setAttribute(qt.Qt.WA_DeleteOnClose)

        self.input = QReconsParams()

    def tearDown(self):
        self.h5Editor.close()
        self.h5Editor = None
        gc.collect()
        self.qapp.processEvents()
        super(TestCaseQt, self).tearDown()

    def testPAGANIN(self):
        """As test are 'raw' and based on the values of the FastSetupDefineGlobals
        We are checking the values of FastSetupDefineGlobals by assert
        to make the difference with the 'True' unittest
        """
        widget = self.h5Editor._PaganinWidget
        self.h5Editor.loadReconsParams(self.input)

        # check Mode conbobox
        assert "MODE" in self.input["PAGANIN"].all_params
        assert self.input["PAGANIN"]["MODE"] == PaganinMode.off
        self.assertTrue(widget._qcbpaganin.currentIndex() == 0)
        self.assertTrue(widget._qcbpaganin.currentText() == "off")

        # test alpha/beta
        assert "DB2" in self.input["PAGANIN"].all_params
        assert self.input["PAGANIN"]["DB2"] == 100.0
        self.assertTrue(widget._qleSigmaBeta2.text() == str(100.0))

        # unsharp coefficient
        assert "UNSHARP_COEFF" in self.input["PAGANIN"].all_params
        assert self.input["PAGANIN"]["UNSHARP_COEFF"] == 3.0
        self.assertTrue(widget._unsharp_sigma_coeff.text() == str(3.0))

        # dilate coefficient
        assert "DILATE" in self.input["PAGANIN"].all_params
        assert self.input["PAGANIN"]["DILATE"] == 2
        self.assertTrue(widget._qleDilatation.text() == str(2))

        # high absorption
        assert "MKEEP_BONE" in self.input["PAGANIN"].all_params
        assert self.input["PAGANIN"]["MKEEP_BONE"] is False
        self.assertFalse(widget._qcbKeepBone.isChecked())

    def testPYHST(self):
        """test that the display of the PyHST widget is correct"""
        widget = self.h5Editor._PyHSTWidget
        self.h5Editor.loadReconsParams(self.input)

        # checkversion
        assert "EXE" in self.input["PYHSTEXE"].all_params
        self.assertTrue(widget._qcbPyHSTVersion.currentText() == PyhstRP.OFFV)

        # checkofficialversion
        assert "OFFV" in self.input["PYHSTEXE"].all_params
        assert self.input["PYHSTEXE"]["OFFV"] == PyhstRP.OFFV
        self.assertTrue(widget._qlOfficalVersion.text() == PyhstRP.OFFV)

        # #check verbose
        assert "VERBOSE" in self.input["PYHSTEXE"].all_params
        assert self.input["PYHSTEXE"]["VERBOSE"] == 0
        self.assertFalse(widget._qcbverbose.isChecked())

        # # check output file
        assert "VERBOSE_FILE" in self.input["PYHSTEXE"].all_params
        assert self.input["PYHSTEXE"]["VERBOSE_FILE"] == "pyhst_out.txt"
        self.assertTrue(widget._qleVerboseFile.text() == "pyhst_out.txt")

    def testBeamGEO(self):
        """test that the display of the BeamGeo widget is correct"""
        widget = self.h5Editor._beamGeoWidget
        self.h5Editor.loadReconsParams(self.input)

        # test reconstruction geometry
        assert "TYPE" in self.input["BEAMGEO"].all_params
        assert self.input["BEAMGEO"]["TYPE"] == BeamGeoType.parallel
        self.assertTrue(widget._qcbType.currentText() == "parallel")

        # test source X
        assert "SX" in self.input["BEAMGEO"].all_params
        assert self.input["BEAMGEO"]["SX"] == 0.0
        self.assertTrue(widget._qleSX.text() == str(0.0))

        # test source Y
        assert "SY" in self.input["BEAMGEO"].all_params
        assert self.input["BEAMGEO"]["SY"] == 0.0
        self.assertTrue(widget._qleSY.text() == str(0.0))

        # test source Distance
        assert "DIST" in self.input["BEAMGEO"].all_params
        assert self.input["BEAMGEO"]["DIST"] == 55.0
        self.assertTrue(widget._qleDIST.text() == str(55.0))

    def testFT(self):
        # parameters from the main widget
        widget = self.h5Editor._mainWidget
        self.h5Editor.loadReconsParams(self.input)

        assert "FIXEDSLICE" in self.input["FT"].all_params
        assert self.input["FT"]["FIXEDSLICE"].name == "middle"
        self.assertTrue(widget._qcbSelectSlice.currentText() == "middle")

        assert "VOLOUTFILE" in self.input["FT"].all_params
        assert self.input["FT"]["VOLOUTFILE"] == 0
        self.assertTrue(widget._qcbEDFStack.isChecked())
        self.assertFalse(widget._qcbSingleVolFile.isChecked())

        assert "CORRECT_SPIKES_THRESHOLD" in self.input["FT"].all_params
        assert self.input["FT"]["CORRECT_SPIKES_THRESHOLD"] == 0.04
        self.assertTrue(widget._qleThresholdSpikesRm.text() == str(0.04))

        assert "DO_TEST_SLICE" in self.input["FT"].all_params
        assert self.input["FT"]["DO_TEST_SLICE"] == 1
        self.assertTrue(widget._qcbDoTestSlice.isChecked())

        assert "VOLSELECT" in self.input["FT"].all_params
        assert self.input["FT"]["VOLSELECT"] == VolSelMode.total
        self.assertTrue(widget._qcbVolumeSelection.currentText() == "total")

        assert "VOLSELECTION_REMEMBER" in self.input["FT"].all_params
        assert self.input["FT"]["VOLSELECTION_REMEMBER"] == 0
        self.assertFalse(widget._qcbVolSelRemenber.isChecked())

        assert "RINGSCORRECTION" in self.input["FT"].all_params
        assert self.input["FT"]["RINGSCORRECTION"] == 0
        self.assertFalse(widget._qcbRingsCorrection.isChecked())

        # parameters from display widget
        widget = self.h5Editor._displayWidget
        self.h5Editor.loadReconsParams(self.input)

        assert "SHOWPROJ" in self.input["FT"].all_params
        assert self.input["FT"]["SHOWPROJ"] == 0
        self.assertFalse(widget._qcbShowProj.isChecked())

        assert "SHOWSLICE" in self.input["FT"].all_params
        assert self.input["FT"]["SHOWSLICE"] == 1
        self.assertTrue(widget._qcbShowSlice.isChecked())

        assert "ANGLE_OFFSET_VALUE" in self.input["FT"].all_params
        assert self.input["FT"]["ANGLE_OFFSET_VALUE"] == 0.0
        self.assertTrue(widget._qleAngleOffset.text() == str(0.0))

        # parameters fron expert widget
        widget = self.h5Editor._expertWidget
        self.h5Editor.loadReconsParams(self.input)

        assert "NUM_PART" in self.input["FT"].all_params
        assert self.input["FT"]["NUM_PART"] == 4
        self.assertTrue(widget._getNumericalPart() == 4)

        assert "VERSION" in self.input["FT"].all_params
        assert self.input["FT"]["VERSION"] == "fastomo3 3.2"
        self.assertTrue(widget._qleVersion.text() == "fastomo3 3.2")

        assert "DATABASE" in self.input["FT"].all_params
        assert self.input["FT"]["DATABASE"] == 0
        self.assertFalse(widget._qcbDataBase.isChecked())

        assert "NO_CHECK" in self.input["FT"].all_params
        assert self.input["FT"]["NO_CHECK"] == 0
        self.assertFalse(widget._qcbNocheck.isChecked())

        assert "ZEROOFFMASK" in self.input["FT"].all_params
        assert self.input["FT"]["ZEROOFFMASK"] == 1
        self.assertTrue(widget._qcbZeroRegionMask.isChecked())

        assert "FIXHD" in self.input["FT"].all_params
        assert self.input["FT"]["FIXHD"] == 0
        self.assertFalse(widget._qcbFixHD.isChecked())

    def testFTAxis(self):
        widget = self.h5Editor._axisWidget
        self.h5Editor.loadReconsParams(self.input)

        self.assertTrue("DO_AXIS_CORRECTION" in self.input["FTAXIS"].all_params)
        self.assertEqual(self.input["FTAXIS"]["DO_AXIS_CORRECTION"], True)
        widget._doAxisCorrection.setChecked(False)
        self.assertTrue(self.input["FTAXIS"]["DO_AXIS_CORRECTION"], False)

        self.assertTrue("USE_TOMWER_AXIS" in self.input["FTAXIS"].all_params)
        self.assertEqual(self.input["FTAXIS"]["USE_TOMWER_AXIS"], True)
        widget._useTomwerAxis.setChecked(False)
        self.assertTrue(self.input["FTAXIS"]["USE_TOMWER_AXIS"], False)

        self.assertTrue("TRY_USE_OLD_TOMWER_AXIS" in self.input["FTAXIS"].all_params)
        self.assertEqual(self.input["FTAXIS"]["TRY_USE_OLD_TOMWER_AXIS"], True)
        widget._useOldTomwerAxis.setChecked(False)
        self.assertTrue(self.input["FTAXIS"]["TRY_USE_OLD_TOMWER_AXIS"], False)

        self.assertTrue("HALF_ACQ" in self.input["FT"].all_params)
        self.assertEqual(self.input["FT"]["HALF_ACQ"], False)
        widget._qcbHalfAcq.isChecked()
        self.assertTrue(self.input["FTAXIS"]["TRY_USE_OLD_TOMWER_AXIS"], True)

    def testDKRF(self):
        widget = self.h5Editor._dkRefWidget
        self.h5Editor.loadReconsParams(self.input)

        assert "DARKCAL" in self.input["DKRF"].all_params
        assert self.input["DKRF"]["DARKCAL"] == dkrf_method.average
        self.assertTrue(widget._qcbDKMode.getMode() == dkrf_method.average)

        assert "REFSCAL" in self.input["DKRF"].all_params
        assert self.input["DKRF"]["REFSCAL"] == dkrf_method.median
        self.assertTrue(widget._qcbRefMode.getMode() == dkrf_method.median)

        assert "DKFILE" in self.input["DKRF"].all_params
        assert self.input["DKRF"]["DKFILE"] == "darkend[0-9]{3,4}"
        self.assertTrue(widget._qleDKPattern.text() == "darkend[0-9]{3,4}")

        assert "RFFILE" in self.input["DKRF"].all_params
        assert self.input["DKRF"]["RFFILE"] == "ref*.*[0-9]{3,4}_[0-9]{3,4}"
        self.assertTrue(widget._qleRefsPattern.text() == "ref*.*[0-9]{3,4}_[0-9]{3,4}")

        assert "REFSRMV" in self.input["DKRF"].all_params
        assert self.input["DKRF"]["REFSRMV"] == 0
        self.assertFalse(widget._qcbRmRef.isChecked())

        assert "REFSOVE" in self.input["DKRF"].all_params
        assert self.input["DKRF"]["REFSOVE"] == 0
        self.assertTrue(widget._qcbSkipRef.isChecked())


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestSimplifyH5EditorSave(TestCaseQt):
    """test that h5editor is returning the correct H5 structures"""

    def setUp(self):
        super(TestSimplifyH5EditorSave, self).setUp()
        self.h5Editor = ReconsParamsEditor(None)
        self.h5Editor.setAttribute(qt.Qt.WA_DeleteOnClose)

        self.input = QReconsParams()
        self.h5Editor.loadReconsParams(self.input)

    def tearDown(self):
        self.qapp.processEvents()
        self.h5Editor.close()
        self.h5Editor = None
        gc.collect()
        self.qapp.processEvents()
        super(TestSimplifyH5EditorSave, self).tearDown()

    def testSaving(self):
        # changing some values in BEAMGEO
        assert self.input.to_dict()["BEAMGEO"]["TYPE"] == "p"
        self.input["BEAMGEO"]["TYPE"] = "c"
        # changing some values in PAGANIN
        assert self.input["PAGANIN"]["MKEEP_BONE"] == 0
        self.input["PAGANIN"]["MKEEP_BONE"] = 1
        # changing some values in PYHSTEXE
        assert self.input["PYHSTEXE"]["VERBOSE_FILE"] == "pyhst_out.txt"
        self.input["PYHSTEXE"]["VERBOSE_FILE"] = "pyhst_out2.txt"
        assert self.input["PYHSTEXE"]["OFFV"] == PyhstRP.OFFV
        self.input["PYHSTEXE"]["OFFV"] = "randomString"
        # changing some FT value
        self.input["FT"]["CORRECT_SPIKES_THRESHOLD"] = "Inf"
        self.input["FT"]["RINGSCORRECTION"] = 1
        self.input["FT"]["ANGLE_OFFSET_VALUE"] = 25.0
        self.input["FT"]["FIXEDSLICE"] = "12"
        self.input["FT"]["SHOWSLICE"] = 0

        # changing Paganin values
        self.input["PAGANIN"]["MODE"] = 3

        # changing the DKRF values
        self.input["DKRF"]["REFSCAL"] = "Average"
        self.input["DKRF"]["DARKCAL"] = "None"
        self.input["DKRF"]["RFFILE"] = "toto.*"
        self.input["DKRF"]["REFSRMV"] = 0

        # loading from new values
        self.h5Editor.loadReconsParams(self.input)

        # changing again some values
        self.h5Editor._PaganinWidget._qcbpaganin.setCurrentIndex(1)
        self.h5Editor._PyHSTWidget._qcbverbose.setCheckState(qt.Qt.Unchecked)
        self.h5Editor._beamGeoWidget._qleSY.setText("12.0")
        # add some checking on the FT
        self.assertFalse(
            self.h5Editor._mainWidget._grpThreshold.isChecked()
        )  # this shouldn't be activated

        # saving the values
        savedStructures = self.h5Editor.getStructs()
        self.assertTrue("BEAMGEO" in savedStructures)
        self.assertTrue("PYHSTEXE" in savedStructures)
        self.assertTrue("PAGANIN" in savedStructures)
        self.assertTrue("FT" in savedStructures)
        self.assertTrue("FTAXIS" in savedStructures)
        self.assertTrue(savedStructures["BEAMGEO"]["TYPE"] == "c")
        self.assertTrue(savedStructures["BEAMGEO"]["SY"] == 12.0)

        self.assertTrue(savedStructures["PAGANIN"]["MKEEP_BONE"] == 1)
        self.assertTrue(savedStructures["PAGANIN"]["MODE"] == 1)

        self.assertTrue(savedStructures["PYHSTEXE"]["VERBOSE_FILE"] == "pyhst_out2.txt")
        self.assertTrue(savedStructures["PYHSTEXE"]["OFFV"] == "randomString")
        self.assertTrue(savedStructures["PYHSTEXE"]["VERBOSE"] == 0)

        self.assertTrue(savedStructures["FT"]["CORRECT_SPIKES_THRESHOLD"] == "Inf")
        self.assertTrue(savedStructures["FT"]["RINGSCORRECTION"] == 1)
        self.assertTrue(savedStructures["FT"]["ANGLE_OFFSET"] == 1)
        self.assertEqual(savedStructures["FT"]["FIXEDSLICE"], "12")
        self.assertTrue(savedStructures["FT"]["SHOWSLICE"] == 0)

        self.assertTrue(savedStructures["FTAXIS"]["DO_AXIS_CORRECTION"] == 1)



# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "08/02/2017"

from silx.gui import qt
import logging
import os


_logger = logging.getLogger()


class QScanDialog(qt.QFileDialog):
    """docstring for QFolderDialog"""

    def __init__(self, parent, multiSelection=False):
        qt.QFileDialog.__init__(self, parent)
        self.setNameFilters(
            [
                "HDF5 files (*.h5 *.hdf5 *.nx *.nexus)",
                "Nexus files (*.nx *.nexus)",
                "Any files (*)",
            ]
        )
        self._selected_files = []
        self.setFileMode(qt.QFileDialog.ExistingFiles)

        # self.QDialogButtonBox
        self.multiSelection = multiSelection
        # check if old data directory exists
        idir = False
        f_cd_last = os.environ["HOME"] + "/.octave/mylastpwd.txt"
        if os.path.isfile(f_cd_last) is True:
            with open(f_cd_last, "r") as fcdl:
                d_cd_last = fcdl.readlines()
                d_cd_last = d_cd_last[0][0 : d_cd_last[0].rfind("/")]
                idir = os.path.isdir(d_cd_last)

        if os.environ.get("TOMWER_DEFAULT_INPUT_DIR", None) and os.path.exists(
            os.environ["TOMWER_DEFAULT_INPUT_DIR"]
        ):
            self.setDirectory(os.environ["TOMWER_DEFAULT_INPUT_DIR"])
        elif idir is True:
            self.setDirectory(d_cd_last)
        elif os.path.isdir("/data"):
            self.setDirectory("/data")

        btns = self.findChildren(qt.QPushButton)
        if len(btns) == 0:
            _logger.error(
                "Cannot retrieve open button. switch to none " "native QFileDialog"
            )
            self.setOption(qt.QFileDialog.DontUseNativeDialog)
            btns = self.findChildren(qt.QPushButton)

        if self.multiSelection is True:
            # to make it possible to select multiple directories:
            self.file_view = self.findChild(qt.QListView, "listView")
            if self.file_view:
                self.file_view.setSelectionMode(qt.QAbstractItemView.MultiSelection)
                self.file_view.setSelectionMode(qt.QAbstractItemView.ExtendedSelection)

            self.f_tree_view = self.findChild(qt.QTreeView)
            if self.f_tree_view:
                self.f_tree_view.setSelectionMode(qt.QAbstractItemView.MultiSelection)
                self.f_tree_view.setSelectionMode(
                    qt.QAbstractItemView.ExtendedSelection
                )

            if len(btns) > 0:
                self.openBtn = [x for x in btns if "open" in str(x.text()).lower()][0]
                self.openBtn.clicked.disconnect()
                self.openBtn.hide()
                parent = self.openBtn.parent()
                self.openBtn = qt.QPushButton("Select", parent=parent)
                self.openBtn.clicked.connect(self.openClicked)
                parent.layout().insertWidget(0, self.openBtn)

    def openClicked(self):
        inds = self.f_tree_view.selectionModel().selectedIndexes()
        for i in inds:
            if i.column() == 0:
                self._selected_files.append(
                    os.path.join(str(self.directory().absolutePath()), str(i.data()))
                )
        self.accept()
        self.done(1)

    def files_selected(self):
        for file_ in self.selectedFiles():
            self._selected_files.append(file_)
        return self._selected_files



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "24/01/2017"


import logging
import os
import shutil
import tempfile
import time
from glob import glob
from silx.gui.utils.testutils import TestCaseQt
from silx.gui import qt
from orangecontrib.tomwer.widgets.control.DataTransfertOW import DataTransfertOW
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils.scanutils import MockEDF

logging.disable(logging.INFO)


class TestEDFFolderTransfertWidget(TestCaseQt):
    """class testing the DataTransfertOW within an EDF acquisition"""

    def setUp(self):
        TestCaseQt.setUp(self)
        self.n_file = 10
        self.sourcedir = tempfile.mkdtemp()
        assert os.path.isdir(self.sourcedir)
        self.targettedir = tempfile.mkdtemp()
        assert os.path.isdir(self.targettedir)
        MockEDF.fastMockAcquisition(self.sourcedir, n_radio=self.n_file)
        self.scan = ScanFactory.create_scan_object(self.sourcedir)

        self.folderTransWidget = DataTransfertOW()
        self.folderTransWidget.turn_off_print = True
        self.folderTransWidget.setDestDir(self.targettedir)
        self.folderTransWidget._copying = False
        self.folderTransWidget.setForceSync(True)

    def tearDown(self):
        self.folderTransWidget.settingsHandler.removeCallback(
            self.folderTransWidget._updateSettingsVals
        )
        self.folderTransWidget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.folderTransWidget.close()
        self.folderTransWidget = None

        if os.path.isdir(self.sourcedir):
            shutil.rmtree(self.sourcedir)
        if os.path.isdir(self.targettedir):
            shutil.rmtree(self.targettedir)
        TestCaseQt.tearDown(self)

    def testMoveFiles(self):
        """
        simple test that files are correctly moved
        """
        self.folderTransWidget._process(self.scan, move=True, noRsync=True)

        outputdir = os.path.join(self.targettedir, os.path.basename(self.sourcedir))
        timeout = 1
        while (
            (os.path.isdir(self.sourcedir))
            and timeout > 0
            or self.folderTransWidget.isCopying()
        ):
            timeout = timeout - 0.1
            time.sleep(0.1)
            self.qapp.processEvents()

        self.assertTrue(os.path.isdir(outputdir))
        self.assertTrue(self.checkDataCopied())

    def testCopyFiles(self):
        """
        Simple test that file are copy and deleted
        """
        self.folderTransWidget._process(self.scan, move=False, noRsync=True)

        timeout = 1
        outputdir = os.path.join(self.targettedir, os.path.basename(self.sourcedir))
        while (
            os.path.isdir(self.sourcedir) and timeout > 0
        ) or self.folderTransWidget.isCopying():
            timeout = timeout - 0.1
            time.sleep(0.1)
            self.qapp.processEvents()

        time.sleep(1)

        self.assertTrue(os.path.isdir(outputdir))
        self.assertTrue(self.checkDataCopied())

    def checkDataCopied(self):
        outputdir = os.path.join(self.targettedir, os.path.basename(self.sourcedir))
        outputFiles = os.listdir(outputdir)
        inputFile = glob(self.sourcedir)

        return (
            (len(inputFile) == 0)
            and (len(outputFiles) == (self.n_file + 3))
            and (not os.path.isdir(self.sourcedir))
        )



# coding: utf-8
###########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

"""
This module is dedicated to cast of volume from one file format to the other.
For now this handle:

* to 16 bits tiff
"""

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "15/12/2021"


import os
import glob
from typing import Optional
from processview.core.superviseprocess import SuperviseProcess
from silx.io.utils import get_data
from tomwer.core.process.task import Task
from silx.utils.enum import Enum as _Enum
from tomwer.core.process.reconstruction.nabu.nabucommon import (
    NabuOutputFileFormat,
    get_file_format,
)
from nabu.io.writer import TIFFWriter
from tomwer.core.process.reconstruction.nabu.nabucommon import _NabuBaseReconstructor
from silx.io.url import DataUrl
from processview.core.manager import ProcessManager, DatasetState
import numpy
import logging
import pathlib
from tomoscan.io import HDF5File
from tomwer.io.utils.utils import get_slice_data

_logger = logging.getLogger(__name__)


class DataType(_Enum):
    UINT_16 = "uint 16"
    UINT_8 = "uint 8"


DATA_SIZE_TO_DTYPE = {
    DataType.UINT_16: numpy.uint16,
    DataType.UINT_8: numpy.uint8,
}


RESCALE_MIN_PERCENTILE = 10
RESCALE_MAX_PERCENTILE = 90

DEFAULT_OUTPUT_DIR = "{scan_basename}/cast_volume"


class CastVolumeTask(
    Task,
    SuperviseProcess,
    input_names=(
        "data",
        "configuration",
    ),
    output_names=("data",),
):
    def __init__(self, process_id=None, varinfo=None, inputs=None, name=None):
        SuperviseProcess.__init__(self, process_id=process_id)
        Task.__init__(self, varinfo=varinfo, inputs=inputs, label=name)

    def run(self):
        scan = self.inputs.data
        configuration = self.inputs.configuration
        output_data_type = DataType.from_value(configuration.get("output_data_type"))
        output_file_format = NabuOutputFileFormat.from_value(
            configuration.get("output_file_format")
        )
        output_file_path = configuration.get("output_file_path", DEFAULT_OUTPUT_DIR)
        # convert some pattern like "{scan_basename}"
        output_file_path = _NabuBaseReconstructor.format_output_location(
            location=output_file_path, scan=scan
        )

        output_data_path = configuration.get("output_data_path", None)
        overwrite = configuration.get("overwrite", False)
        rescale_min_percentile = configuration.get(
            "rescale_min_percentile", RESCALE_MIN_PERCENTILE
        )
        rescale_max_percentile = configuration.get(
            "rescale_max_percentile", RESCALE_MAX_PERCENTILE
        )
        if scan.latest_vol_reconstructions is None:
            mess = f"volume cast of {str(scan)} skipped. No volume found in latest reconstruction"
            _logger.processSkipped(mess)
            state = DatasetState.SKIPPED
        else:
            try:
                cast_volumes(
                    input_volume_urls=scan.latest_vol_reconstructions,
                    output_data_type=output_data_type,
                    output_file_format=output_file_format,
                    output_file_path=output_file_path,
                    output_data_path=output_data_path,
                    overwrite=overwrite,
                    scan=scan,
                    rescale_min_percentile=rescale_min_percentile,
                    rescale_max_percentile=rescale_max_percentile,
                )
            except Exception as e:
                mess = f"volume cast of {str(scan)} failed. Reason is {str(e)}"
                state = DatasetState.FAILED
            else:
                mess = f"volume cast of {str(scan)} succeed"
                state = DatasetState.SUCCEED

            _logger.processSkipped(mess)
            ProcessManager().notify_dataset_state(
                dataset=scan,
                process=self,
                state=state,
            )


class Cast:
    def __init__(
        self,
        input_file_format: NabuOutputFileFormat,
        output_data_type: DataType,
        output_file_format: NabuOutputFileFormat,
    ) -> None:
        self._input_file_format = input_file_format
        self._output_data_type = output_data_type
        self._output_file_format = output_file_format

    @property
    def input_file_format(self):
        return self._input_file_format

    @property
    def output_data_type(self):
        return self._output_data_type

    @property
    def output_file_format(self):
        return self._output_file_format

    def __str__(self):
        return f"""cast: \n
        - inputs: file format {self.input_file_format.value}
        - outputs: file format {self.input_file_format.value}, data size {self.output_data_type}
        """

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Cast):
            return False
        else:
            return (
                self.input_file_format == other.input_file_format
                and self.output_data_type == other.output_data_type
                and self.output_file_format == other.output_file_format
            )


HANDLED_CASTS = (
    Cast(
        NabuOutputFileFormat.TIFF,
        DataType.UINT_16,
        output_file_format=NabuOutputFileFormat.TIFF,
    ),
    Cast(
        NabuOutputFileFormat.TIFF,
        DataType.UINT_8,
        output_file_format=NabuOutputFileFormat.TIFF,
    ),
    Cast(
        NabuOutputFileFormat.HDF5,
        DataType.UINT_16,
        output_file_format=NabuOutputFileFormat.TIFF,
    ),
    Cast(
        NabuOutputFileFormat.HDF5,
        DataType.UINT_8,
        output_file_format=NabuOutputFileFormat.TIFF,
    ),
)


def rescale_data(
    data: numpy.ndarray,
    dtype,
    new_min,
    new_max,
    data_min=None,
    data_max=None,
    rescale_min_percentile=RESCALE_MIN_PERCENTILE,
    rescale_max_percentile=RESCALE_MAX_PERCENTILE,
):
    if data_min is None or data_max is None:
        try:
            from silx.math.combo import min_max

            data_min, data_max = min_max(data, finite=True)
        except Exception:
            data_min = numpy.percentile(data, rescale_min_percentile)
            data_max = numpy.percentile(data, rescale_max_percentile)
    # in order to fit to percentiles
    data[data < data_min] = data_min
    data[data > data_max] = data_max
    data = ((data - data_min) / (data_max - data_min)) * (new_max - new_min) + new_min
    return data.astype(dtype)


def find_histogram(url: DataUrl, scan=None) -> tuple:
    """
    Look for histogram of the provided url. If found one return the DataUrl of the nabu histogram
    """
    if not isinstance(url, DataUrl):
        raise TypeError(
            f"url is expected to be an instance of {DataUrl} not {type(url)}"
        )
    else:
        file_dir = os.path.realpath(os.path.dirname(url.file_path()))
        histo = glob.glob(os.path.join(file_dir, "*histogram.hdf5"))
        if len(histo) >= 1:
            if len(histo) > 1:
                _logger.warning(
                    f"more than one histogram found ({histo}). Pick the first one"
                )
            file_path = os.path.join(file_dir, histo[0])
            if scan is None:
                # take the first entry found
                with HDF5File(file_path, mode="r") as h5s:
                    keys = list(h5s.keys())

                if len("keys") == 0:
                    _logger.error(
                        f"No entry found in {file_path}. Unable to find histogram"
                    )
                    return None
                else:
                    entry = keys[0]
                    if len(keys) > 1:
                        _logger.warning(
                            f"more than one entry found in {file_path}. Take the first one {keys[0]}"
                        )
            else:
                entry = scan.entry if hasattr(scan, "entry") else "entry"
            return DataUrl(
                file_path=file_path,
                data_path=f"/{entry}/histogram/results/data",
                scheme="silx",
            )
        else:
            _logger.info(f"no histogram foudn at {file_dir}")
            return None


def _get_hst_saturations(hist, bins, rescale_min_percentile, rescale_max_percentile):
    hist_cum = numpy.cumsum(hist)
    bin_index_min = numpy.searchsorted(
        hist_cum, numpy.percentile(hist_cum, rescale_min_percentile)
    )
    bin_index_max = numpy.searchsorted(
        hist_cum, numpy.percentile(hist_cum, rescale_max_percentile)
    )
    return bins[bin_index_min], bins[bin_index_max]


def _try_to_find_min_max_from_histo(
    input_volume_url, rescale_min_percentile, rescale_max_percentile, scan=None
) -> tuple:
    histogram_res_url = find_histogram(input_volume_url, scan=scan)
    data_min = data_max = None
    if histogram_res_url is not None:
        try:
            histogram = get_data(histogram_res_url)
        except Exception as e:
            _logger.error(
                f"Fail to load histogram from {histogram_res_url}. Reason is {e}"
            )
        else:
            bins = histogram[1]
            hist = histogram[0]
            data_min, data_max = _get_hst_saturations(
                hist, bins, rescale_min_percentile, rescale_max_percentile
            )
    return data_min, data_max


def cast_volumes(
    input_volume_urls: tuple,
    output_data_type: DataType,
    output_file_format: NabuOutputFileFormat,
    output_file_path: str,
    output_data_path: Optional[str] = None,
    overwrite=False,
    data_min=None,
    data_max=None,
    scan=None,
    rescale_min_percentile=RESCALE_MIN_PERCENTILE,
    rescale_max_percentile=RESCALE_MAX_PERCENTILE,
) -> tuple:
    """

    :param input_volume_urls: Iterable of DataUrl pointing to the volume to convert.
                              It expect all urls to be associated to the same volume
                              in order to compute min/max on it and rescale data.
    :param DataType output_data_type: the data type you want in output
    :param NabuOutputFileFormat output_file_format: output file format
    :param str output_file_path: path to save the result.
    :param str output_data_path: output data path for hdf5 files
    :param bool overwrite: overwrite file / dataset if already exists
    :return: tuple of created DataUrl
    :rtype: tuple
    """
    if not os.path.exists(output_file_path):
        os.makedirs(output_file_path, exist_ok=True)

    if len(input_volume_urls) == 0:
        return

    if data_min is None or data_max is None:
        data_min, data_max = _try_to_find_min_max_from_histo(
            input_volume_url=input_volume_urls[0],
            scan=scan,
            rescale_min_percentile=rescale_min_percentile,
            rescale_max_percentile=rescale_max_percentile,
        )
    # if failed to find histogram from nabu histogram file then recompute it
    if data_min is None or data_max is None:
        data_min = numpy.finfo(numpy.float64).max
        data_max = numpy.finfo(numpy.float64).min
        for input_volume_url in input_volume_urls:
            data = get_slice_data(input_volume_url)
            data_min = min(data_min, numpy.min(data))
            data_max = max(data_max, numpy.max(data))

    res = []
    for input_volume_url in input_volume_urls:
        res.append(
            cast_volume(
                input_volume_url=input_volume_url,
                output_data_type=output_data_type,
                output_file_format=output_file_format,
                output_file_path=output_file_path,
                output_data_path=output_data_path,
                overwrite=overwrite,
                data_min=data_min,
                data_max=data_max,
            )
        )
    return tuple(res)


def cast_volume(
    input_volume_url: DataUrl,
    output_data_type: DataType,
    output_file_format: NabuOutputFileFormat,
    output_file_path: str,
    output_data_path: Optional[str] = None,
    overwrite=False,
    data_min=None,
    data_max=None,
    scan=None,
    rescale_min_percentile=RESCALE_MIN_PERCENTILE,
    rescale_max_percentile=RESCALE_MAX_PERCENTILE,
) -> DataUrl:
    if not isinstance(input_volume_url, DataUrl):
        raise TypeError(
            f"first parameter 'input_volume_url' is expected to be a DataUrl not {type(input_volume_url)}"
        )
    output_data_type = DataType.from_value(output_data_type)
    output_file_format = NabuOutputFileFormat.from_value(output_file_format)
    if not isinstance(output_file_path, (str, pathlib.Path)):
        raise TypeError(
            f"output_file_path is expected to be a str not {type(output_file_path)}"
        )
    if output_data_path is not None and not isinstance(
        output_data_path, (str, pathlib.Path)
    ):
        raise TypeError(
            f"output_data_path is expected to be a str not {type(output_data_path)}"
        )

    if not isinstance(overwrite, bool):
        raise TypeError(f"overwrite is expected to be a bool not {type(overwrite)}")

    cast = Cast(
        input_file_format=get_file_format(input_volume_url.file_path()),
        output_data_type=output_data_type,
        output_file_format=output_file_format,
    )
    if cast not in HANDLED_CASTS:
        handled_cast = "\n".join([str(c_) for c_ in HANDLED_CASTS])
        raise ValueError(
            f"cast {str(cast)} \n is not handled. Handled cast are {handled_cast} \n"
        )
    if data_min is None or data_max is None:
        data_min, data_max = _try_to_find_min_max_from_histo(
            input_volume_url=input_volume_url, scan=scan
        )
    data = get_slice_data(input_volume_url)
    if output_data_type is DataType.UINT_16:
        data = rescale_data(
            data=data,
            new_min=numpy.iinfo(numpy.uint16).min,
            new_max=numpy.iinfo(numpy.uint16).max,
            dtype=numpy.uint16,
            data_min=data_min,
            data_max=data_max,
            rescale_min_percentile=rescale_min_percentile,
            rescale_max_percentile=rescale_max_percentile,
        )
        data = data.astype(dtype=DATA_SIZE_TO_DTYPE[output_data_type])
    elif output_data_type is DataType.UINT_8:
        data = rescale_data(
            data=data,
            new_min=numpy.iinfo(numpy.uint8).min,
            new_max=numpy.iinfo(numpy.uint8).max,
            dtype=numpy.uint8,
            data_min=data_min,
            data_max=data_max,
            rescale_min_percentile=rescale_min_percentile,
            rescale_max_percentile=rescale_max_percentile,
        )
        data = data.astype(dtype=DATA_SIZE_TO_DTYPE[output_data_type])
    else:
        raise ValueError(f"{output_data_type} is not handled")

    # deduce output file path
    # for now we can only save to a directory under the same name or
    # to a file for which the path is provided already
    if os.path.isdir(output_file_path):
        original_name = os.path.basename(input_volume_url.file_path())
        original_name_with_new_ext = ".".join(
            [os.path.splitext(original_name)[0], output_file_format.value]
        )
        output_file_path = os.path.join(output_file_path, original_name_with_new_ext)
    if os.path.exists(output_file_path):
        if overwrite:
            os.remove(output_file_path)
        else:
            raise OSError("output file alrady exists ({output_file_path})")
    if output_file_format is NabuOutputFileFormat.TIFF:
        tiff_writer = TIFFWriter(output_file_path)
        tiff_writer.write(data)
        return DataUrl(
            file_path=output_file_path, scheme="cv2"
        )  # use cv2 because fabio seems to be unhappy to read float16 dataset.
    else:
        raise ValueError(f"{output_file_format} is not handled")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Utils for devices (GPU, CPU...)"""

__authors__ = ["H. Payno"]

try:
    import pycuda.driver as cuda
except ImportError:
    cuda = None
from silx.opencl import ocl


class _DeviceBase:
    """base class for device definition"""

    def __init__(self, name):
        self._name = name

    @property
    def name(self):
        return self._name


class OpenCLDevice(_DeviceBase):
    """Definition of an opencl device"""

    def __init__(self, name, platform_id, device_id):
        super(OpenCLDevice, self).__init__(name=name)
        self._platform_id = platform_id
        self._device_id = device_id

    @property
    def platform_id(self):
        return self._platform_id

    @property
    def device_id(self):
        return self._device_id


class CudaDevice(_DeviceBase):
    """Definition of an opencl device"""

    def __init__(self, name, id_):
        super(CudaDevice, self).__init__(name=name)
        self._id = id_

    @property
    def id(self):
        """

        :return: ID of the cuda device
        :rtype: int
        """
        return self._id


class _CudaPlatformBase:
    def getExistingDevices(self):
        """

        :return: all existing cuda device
        """
        devices = []
        if cuda is not None:
            cuda.init()
            for i_device in range(cuda.Device.count()):
                device = CudaDevice(cuda.Device(i_device).name(), id_=i_device)
                devices.append(device)
        return devices


class _OpenCLPlatformBase:
    def getExistingDevices(self):
        """

        :return: all existing opencl platform
        """
        devices = []
        if ocl is not None:
            for platform_id, platform in enumerate(ocl.platforms):
                for device_id, dev in enumerate(platform.devices):
                    devices.append(
                        OpenCLDevice(
                            name=dev.name, platform_id=platform_id, device_id=device_id
                        )
                    )
        return devices
